from abc import abstractmethod, ABC
from random import shuffle
from typing import Dict, Iterator, List, Optional, Sequence, Union
from itertools import cycle
from copy import copy
from warnings import warn
from numpy import nditer, argmin
from numpy import array as nparray
from math import floor
import numpy as np


from hub.constants import MB
from hub.core.chunk.base_chunk import BaseChunk
from hub.core.chunk_engine import ChunkEngine
from hub.core.linked_chunk_engine import LinkedChunkEngine
from hub.core.meta.encode.base_encoder import LAST_SEEN_INDEX_COLUMN
from hub.core.meta.encode.chunk_id import CHUNK_ID_COLUMN, ChunkIdEncoder
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.storage import LRUCache, MemoryProvider, StorageProvider, LocalProvider
from hub.core.tiling.deserialize import combine_chunks
from hub.util.exceptions import (
    DatasetUnsupportedPytorch,
    SampleDecompressionError,
)
from hub.util.keys import get_chunk_key, get_tensor_meta_key
from hub.util.remove_cache import get_base_storage
from hub.util.storage import get_pytorch_local_storage


ChunkEngineMap = Dict[str, ChunkEngine]
CachesMap = Dict[str, LRUCache]


class IOBlock:
    """
    Represents ordered sequential read of samples from corresponding tensor chunks.
    """

    def __init__(self, chunks: List[List[Optional[str]]], indexes: List[int]) -> None:
        self._chunks: List[List[Optional[str]]] = chunks
        self._ind: List[int] = indexes

    def shuffle(self):
        r"""
        Shuffle sequence in which indices would be read from the IOBlock
        """
        shuffle(self._ind)

    def chunk_names(self, tensor_index: int) -> List[Optional[str]]:
        return self._chunks[tensor_index]

    def indices(self) -> List[int]:
        return self._ind

    def chunks(self) -> List[List[Optional[str]]]:
        return self._chunks

    def split(self, n) -> List["IOBlock"]:
        k, m = divmod(len(self._ind), n)
        return [
            IOBlock(
                self._chunks, self._ind[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)]
            )
            for i in range(n)
        ]

    def __len__(self) -> int:
        return len(self._ind)


class Schedule:
    def __init__(self, blocks: List[IOBlock]) -> None:
        self._blocks: List[IOBlock] = blocks

    def shuffle(self) -> None:
        r"""
        Shuffle IOBlocks in the schedule as well as each IOBlock
        """
        shuffle(self._blocks)

        for block in self._blocks:
            block.shuffle()

    def __iter__(self):
        return iter(self._blocks)

    def __len__(self):
        return sum(map(len, self._blocks))


class Scheduler(ABC):
    @abstractmethod
    def schedule(self, jobs: List[IOBlock]) -> List[Schedule]:
        ...


class SingleThreadScheduler(Scheduler):
    def schedule(self, jobs: List[IOBlock]) -> List[Schedule]:
        return [Schedule(jobs)]


class SequentialMultithreadScheduler(Scheduler):
    """
    Splits list of IO blocks in a way, so PyTorch loader would return
    samples in sequence, when started with `num_worker` > 1.

    Scheduler relays on a fact, that PyTorch DataLoader synchronize
    read of samples per thread and return in a sequence per worker.

    Example:
        Given sequence of indices `[1, 2, 3, 4, 5, 6]` and 4 workers
        PyTorch have to read samples in an order
            thread0: [1, 5]
            thread1: [2, 6]
            thread2: [3]
            thread3: [4]

        So that initial order would be reconstructed by the DataLoader
    """

    def __init__(self, num_workers: int) -> None:
        super().__init__()
        self.num_workers = num_workers

    def schedule(self, jobs: List[IOBlock]) -> List[Schedule]:
        per_worker: List[List[IOBlock]] = [list() for _ in range(self.num_workers)]
        assigned_worker = iter(cycle(range(self.num_workers)))

        for job in jobs:
            split: List[List[int]] = [list() for _ in range(self.num_workers)]

            for ind in job.indices():
                split[next(assigned_worker)].append(ind)

            for worker_id, idx_list in enumerate(split):
                if len(idx_list) > 0:
                    worker_block = IOBlock(job.chunks(), idx_list)
                    per_worker[worker_id].append(worker_block)

        return [Schedule(worker_jobs) for worker_jobs in per_worker]


class MultiThreadedNaiveScheduler(Scheduler):
    def __init__(self, num_workers: int) -> None:
        super().__init__()
        self.num_workers = num_workers

    def schedule(self, jobs: List[IOBlock]) -> List[Schedule]:
        return [Schedule(blocks) for blocks in self.split(jobs, self.num_workers)]

    def split(self, inlist, n):
        k, m = divmod(len(inlist), n)
        return (
            inlist[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n)
        )


class ShufflingSchedulerWrapper(Scheduler):
    def __init__(self, other: Scheduler) -> None:
        super().__init__()
        self.other: Scheduler = other

    def schedule(self, jobs: List[IOBlock]) -> List[Schedule]:
        schedules = self.other.schedule(jobs)
        for schedule in schedules:
            schedule.shuffle()

        return schedules


class DistributedScheduler(Scheduler):
    """Scheduler arrange IOBlocks between multiple processes and ensure equal
    distribution for each. Initial `List[IOBlock]` order is preserved.
    """

    def __init__(self, num_worker: int = 0) -> None:
        super().__init__()
        self.next_scheduler: Optional[Scheduler] = (
            MultiThreadedNaiveScheduler(num_worker) if num_worker > 0 else None
        )

    def schedule(self, jobs: List[IOBlock]) -> List[Schedule]:
        import torch.distributed as dist
        import torch

        assert dist.is_available()
        assert dist.is_initialized()

        world_size = dist.get_world_size()
        rank = dist.get_rank()

        gr = dist.new_group([i for i in range(world_size)], backend="gloo")

        blocks_len: torch.Tensor = torch.tensor([len(j.indices()) for j in jobs])
        all_idx: torch.Tensor = torch.zeros(
            (sum([len(j.indices()) for j in jobs]), 2), dtype=torch.int
        )

        if rank == 0:
            all_idx = all_idx[0 : floor(len(all_idx) / world_size) * world_size, :]
            all_idx[:, 0] = torch.repeat_interleave(
                torch.arange(len(jobs)), blocks_len
            )[: len(all_idx)]
            all_idx[:, 1] = torch.tensor(
                [i for j in jobs for i in j.indices()][: len(all_idx)]
            )

        thread_local_idx: torch.Tensor = torch.zeros(
            (int(len(all_idx) / world_size), 2), dtype=torch.int
        )

        dist.scatter(
            thread_local_idx,
            scatter_list=list(all_idx.chunk(world_size)) if rank == 0 else None,
            src=0,
            group=gr,
        )

        # recombine assigned blocks
        blocks_map: Dict[int, List[int]] = dict()
        for idx in thread_local_idx:
            key = int(idx[0])
            val = int(idx[1])

            if key in blocks_map:
                blocks_map[key].append(val)
            else:
                blocks_map[key] = [val]

        blocks = [IOBlock(jobs[k].chunks(), v) for k, v in blocks_map.items()]

        if self.next_scheduler:
            return self.next_scheduler.schedule(blocks)
        else:
            return [Schedule(blocks)]


class Streaming(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def read(self, schedule: Schedule) -> Iterator:
        r"""
        Args:
            schedule(Schedule) schedule of IOBlocks to stream
        Returns:
            generator over specific Schedule
        """
        ...


class SampleStreaming(Streaming):
    def __init__(
        self,
        dataset,
        tensors: Sequence[str],
        tobytes: Union[bool, Sequence[str]] = False,
        use_local_cache: bool = False,
        return_index: bool = True,
        pad_tensors: bool = False,
    ) -> None:
        super().__init__()

        self.dataset = dataset
        self.local_storage: Optional[LocalProvider] = (
            get_pytorch_local_storage(dataset) if use_local_cache else None
        )

        # TODO: copy all meta/info to local_storage
        self.storage = get_base_storage(dataset.storage)
        if isinstance(self.storage, MemoryProvider):
            raise DatasetUnsupportedPytorch(
                "The underlying storage is MemoryProvider which isn't supported."
            )

        self.tensors = tensors
        self.pad_tensors = pad_tensors
        if isinstance(tobytes, bool):
            self.tobytes = {k: tobytes for k in self.tensors}
        else:
            for k in tobytes:
                if k not in tensors:
                    raise Exception(
                        f"Tensor {k} is not present in the list of provided tensors: {tensors}."
                    )
            self.tobytes = {k: k in tobytes for k in tensors}

        self.chunk_engines: ChunkEngineMap = self._map_chunk_engines(self.tensors)

        self.local_caches: Optional[CachesMap] = (
            ({tensor: self._use_cache(self.local_storage) for tensor in self.tensors})
            if self.local_storage is not None
            else None
        )

        self.return_index = return_index

    def read(self, schedule: Schedule) -> Iterator:
        for block in schedule._blocks:
            yield from self.stream(block)

    def stream(self, block: IOBlock):
        for idx in block.indices():

            sample = dict()
            valid_sample_flag = True

            for keyid, (key, engine) in enumerate(self.chunk_engines.items()):
                decompress = not self.tobytes[key]
                chunk_class = engine.chunk_class
                try:

                    chunks: List[BaseChunk] = []
                    c_names = block.chunk_names(keyid)

                    version_state = self.dataset.version_state
                    if c_names == [None]:
                        sample[key] = engine.get_empty_sample()
                        continue
                    for c_name in c_names:
                        commit_id = engine.get_chunk_commit(c_name)
                        c_key = get_chunk_key(
                            version_state["tensor_names"][key],
                            c_name,  # type: ignore
                            commit_id,
                        )
                        if self.local_caches is not None:
                            local_cache = self.local_caches[key]

                            if c_key in local_cache:
                                chunk = local_cache.get_hub_object(c_key, chunk_class, meta=engine.chunk_args)  # type: ignore
                            else:
                                chunk = engine.get_chunk(c_key)
                                local_cache[c_key] = chunk

                                # send data to actual storage
                                local_cache._forward(c_key)
                        else:
                            chunk = engine.get_chunk(c_key)
                        chunks.append(chunk)
                    if len(chunks) == 1:
                        data = engine.read_sample_from_chunk(
                            idx, chunk, decompress=decompress
                        )
                    else:
                        if not decompress:
                            raise NotImplementedError(
                                "`tobytes=True` is not supported by tiled samples as it can cause recompression."
                            )
                        data = combine_chunks(chunks, idx, engine.tile_encoder)

                    if data is not None:
                        sample[key] = data
                    else:
                        valid_sample_flag = False
                        break
                except SampleDecompressionError:
                    warn(
                        f"Skipping corrupt {engine.tensor_meta.sample_compression} sample at dataset.{key}[{idx}]"
                    )
                    valid_sample_flag = False
                    break

            if valid_sample_flag:
                if self.return_index:
                    sample["index"] = np.array([idx])
                yield sample

    def _get_block_for_single_sample(self, idx):
        chunks = []
        for engine in self.chunk_engines.values():
            enc = engine.chunk_id_encoder
            try:
                cids = enc[idx]
                cnames = map(enc.name_from_id, cids)
                chunks.append(cnames)
            except Exception:
                chunks.append([None])
        return IOBlock(chunks, [idx])

    def _is_continuious(self):
        idx_entry = self.dataset.index.values[0]
        if isinstance(idx_entry.value, slice):
            step = idx_entry.value.step
            if step and step < 1:
                return False
            return True
        else:
            prev = -1
            for idx in idx_entry.indices(self._get_dataset_length()):
                if idx < prev:
                    return False
                prev = idx
            return True

    def list_blocks(self) -> List[IOBlock]:
        if self._is_continuious():
            return self.list_blocks_continuous()
        return self.list_blocks_random()

    def list_blocks_random(self) -> List[IOBlock]:
        return list(map(self._get_block_for_single_sample, self._get_dataset_indices()))

    def _intersection(self, index, low, high):
        if isinstance(index, slice):
            start = index.start or 0
            stop = index.stop or self._get_dataset_length()
            step = index.step
            if step is None or step == 1:

                return list(range(max(start, low), min(stop, high)))
            else:
                if start < low:
                    rm = (low - start) % step
                    start = low + bool(rm) * (step - rm)
                return list(range(start, min(stop, high), step))
        elif isinstance(index, (list, tuple)):
            return index[np.searchsorted(index, low) : np.searchsorted(index, high)]
        elif isinstance(index, int):
            return [index] if index >= low and index < high else []
        else:
            raise TypeError(index)

    def list_blocks_continuous(self) -> List[IOBlock]:
        blocks: List[IOBlock] = list()

        chunk_id_encodings = [
            engine.chunk_id_encoder.array for engine in self.chunk_engines.values()
        ]

        iterators = [
            nditer([arr[:, LAST_SEEN_INDEX_COLUMN], arr[:, CHUNK_ID_COLUMN]])  # type: ignore
            for arr in chunk_id_encodings
        ]

        last_idx: int = 0

        check_fn = any if self.pad_tensors else all
        while check_fn([not it.finished for it in iterators]):
            next_it = iterators[
                argmin(nparray([it.value[0] for it in iterators if not it.finished]))
            ]
            next_it_value = int(next_it.value[0])

            if next_it_value >= last_idx:
                chunks: List[List[Optional[str]]] = []
                for it in iterators:
                    if it.finished:
                        chunks.append([None])
                        continue
                    cur_ids = []
                    if it.value[0] == next_it_value:
                        while not it.finished and it.value[0] == next_it_value:
                            cur_ids.append(it.value[1])
                            it.iternext()
                    else:
                        cur_ids.append(it.value[1])
                    cur_chunks: List[Optional[str]] = [
                        ChunkIdEncoder.name_from_id(cid)  # type: ignore
                        for cid in cur_ids
                    ]
                    chunks.append(cur_chunks)

                streamable_ids = self._intersection(
                    self.dataset.index.values[0].value, last_idx, next_it_value + 1
                )

                if streamable_ids:
                    new_block = IOBlock(chunks, streamable_ids)
                    blocks.append(new_block)

                last_idx = next_it_value + 1

        return blocks

    def _use_cache(self, storage: Union[StorageProvider, LRUCache]) -> LRUCache:
        cache = LRUCache(MemoryProvider(), copy(storage), 32 * MB)
        cache.read_only = storage.read_only
        return cache

    def _map_chunk_engines(self, tensors: Sequence[str]) -> Dict[str, ChunkEngine]:
        return {
            name: self._create_chunk_engine(name, self.dataset.version_state)
            for name in tensors
        }

    def _create_chunk_engine(self, tensor_name, version_state):
        tensor_key = version_state["tensor_names"][tensor_name]
        meta_key = get_tensor_meta_key(tensor_key, version_state["commit_id"])
        cache = self._use_cache(self.storage)
        meta = cache.get_hub_object(meta_key, TensorMeta)
        if meta.is_link:
            return LinkedChunkEngine(
                tensor_key,
                cache,
                version_state,
                link_creds=self.dataset.link_creds,
            )
        return ChunkEngine(tensor_key, cache, version_state)

    def _get_dataset_length(self):
        version_state = self.dataset.version_state
        tensor_lengths = [
            len(version_state["full_tensors"][version_state["tensor_names"][tensor]])
            for tensor in self.tensors
        ]
        length_fn = max if self.pad_tensors else min
        return length_fn(tensor_lengths, default=0)

    def _get_dataset_indices(self):
        return self.dataset.index.values[0].indices(self._get_dataset_length())



from typing import Optional, Tuple, Union
import numpy as np

from hub.core.compression import compress_array
from hub.core.tiling.optimizer import get_tile_shape
from hub.core.tiling.serialize import break_into_tiles, serialize_tiles, get_tile_shapes
from hub.util.compression import get_compression_ratio
from hub.compression import BYTE_COMPRESSIONS
from hub.constants import MB


class SampleTiles:
    """Stores the tiles corresponding to a sample."""

    def __init__(
        self,
        arr: Optional[np.ndarray] = None,
        compression: Optional[str] = None,
        chunk_size: int = 16 * MB,
        store_uncompressed_tiles: bool = False,
        htype: Optional[str] = None,
        tile_shape: Optional[Tuple[int, ...]] = None,
        sample_shape: Optional[Tuple[int, ...]] = None,
        dtype: Optional[Union[np.dtype, str]] = None,
    ):
        self.registered = False
        self.tiles_yielded = 0
        if arr is not None:
            self._init_from_array(
                arr,
                compression,
                chunk_size,
                store_uncompressed_tiles,
                htype,
                tile_shape,
            )
        else:
            self._init_from_sample_shape(
                sample_shape,  # type: ignore
                compression,
                chunk_size,
                store_uncompressed_tiles,
                htype,
                tile_shape,
                dtype,
            )

    def _init_from_array(
        self,
        arr: np.ndarray,
        compression: Optional[str] = None,
        chunk_size: int = 16 * MB,
        store_uncompressed_tiles: bool = False,
        htype: Optional[str] = None,
        tile_shape: Optional[Tuple[int, ...]] = None,
    ):
        self.arr = arr
        self.sample_shape = arr.shape
        tile_shape = tile_shape or self._get_tile_shape(
            arr.dtype, htype, chunk_size, compression  # type: ignore
        )
        self.tile_shape = tile_shape
        tiles = break_into_tiles(arr, tile_shape)
        self.tiles = serialize_tiles(tiles, lambda x: compress_array(x, compression))
        tile_shapes = np.vectorize(lambda x: x.shape, otypes=[object])(tiles)

        self.shapes_enumerator = np.ndenumerate(tile_shapes)
        self.layout_shape = self.tiles.shape
        self.num_tiles = self.tiles.size
        self.tiles_enumerator = np.ndenumerate(self.tiles)
        self.uncompressed_tiles_enumerator = (
            np.ndenumerate(tiles) if store_uncompressed_tiles else None
        )

    def _init_from_sample_shape(
        self,
        sample_shape: Tuple[int, ...],
        compression: Optional[str] = None,
        chunk_size: int = 16 * MB,
        store_uncompressed_tiles: bool = False,
        htype: Optional[str] = None,
        tile_shape: Optional[Tuple[int, ...]] = None,
        dtype: Optional[Union[np.dtype, str]] = None,
    ):
        self.arr = None  # type: ignore
        self.sample_shape = sample_shape
        self.tiles = None  # type: ignore
        tile_shape = tile_shape or self._get_tile_shape(
            dtype, htype, chunk_size, compression  # type:ignore
        )
        self.tile_shape = tile_shape
        tile_shapes = get_tile_shapes(self.sample_shape, tile_shape)
        self.shapes_enumerator = np.ndenumerate(tile_shapes)
        self.layout_shape = tile_shapes.shape
        self.num_tiles = tile_shapes.size
        self.uncompressed_tiles_enumerator = None

    def _get_tile_shape(
        self, dtype: Union[np.dtype, str], htype: str, chunk_size: int, compression: str
    ):
        # Exclude channels axis from tiling for image, video and audio
        exclude_axis = (
            None
            if htype == "generic"
            and (not compression or compression in BYTE_COMPRESSIONS)
            else -1
        )
        return get_tile_shape(
            self.sample_shape,
            np.prod(np.array(self.sample_shape, dtype=np.uint64))  # type: ignore
            * np.dtype(dtype).itemsize
            * get_compression_ratio(compression),
            chunk_size,
            exclude_axis,
        )

    @property
    def is_first_write(self) -> bool:
        return self.tiles_yielded == 1

    @property
    def is_last_write(self) -> bool:
        return self.tiles_yielded == self.num_tiles

    def yield_tile(self):
        self.tiles_yielded += 1
        if self.tiles is None:
            tile = b""
        else:
            tile = next(self.tiles_enumerator)[1]
        return tile, next(self.shapes_enumerator)[1]

    def yield_uncompressed_tile(self):
        if self.uncompressed_tiles_enumerator is not None:
            return next(self.uncompressed_tiles_enumerator)[1]



import numpy as np

import hub
from hub.core.compression import decompress_array
from hub.core.serialize import serialize_numpy_and_base_types, serialize_sample_object
from hub.constants import KB, MB
from hub.core.tiling.deserialize import np_list_to_sample
from hub.core.tiling.sample_tiles import SampleTiles


def test_numpy_and_base_types():
    arr1 = np.ones((101, 100, 3), dtype=np.int32)
    out1, shape = serialize_numpy_and_base_types(
        arr1, None, None, "int64", "generic", 16 * MB
    )
    arr1_deserialized = np.frombuffer(out1, dtype=np.int64).reshape(shape)
    np.testing.assert_array_equal(arr1, arr1_deserialized)

    out2, shape = serialize_numpy_and_base_types(
        arr1, None, None, "int64", "generic", 100 * KB
    )
    assert isinstance(out2, SampleTiles)
    out_list = [out2.yield_tile() for _ in range(out2.num_tiles)]
    np_list = [np.frombuffer(b[0], dtype=np.int64).reshape(b[1]) for b in out_list]
    tile_shape, layout_shape = out2.tile_shape, out2.layout_shape
    out2 = np_list_to_sample(np_list, shape, tile_shape, layout_shape, "int64")
    np.testing.assert_array_equal(arr1, out2)

    arr2 = np.random.randint(0, 255, (501, 503, 3), dtype="uint8")
    out2, shape = serialize_numpy_and_base_types(
        arr2, "png", None, "uint8", "generic", 16 * MB
    )
    arr2_deserialized = decompress_array(out2, compression="png").reshape(shape)
    np.testing.assert_array_equal(arr2, arr2_deserialized)

    out3, shape = serialize_numpy_and_base_types(
        arr2, "png", None, "uint8", "generic", 100 * KB
    )
    assert isinstance(out3, SampleTiles)
    out_list = [out3.yield_tile() for _ in range(out3.num_tiles)]
    np_list = [
        decompress_array(b[0], compression="png").reshape(b[1]) for b in out_list
    ]
    tile_shape, layout_shape = out3.tile_shape, out3.layout_shape
    out3 = np_list_to_sample(np_list, shape, tile_shape, layout_shape, "uint8")
    np.testing.assert_array_equal(arr2, out3)


def test_sample_img_compression(cat_path, compression="png"):
    sample = hub.read(cat_path)
    arr = sample.array

    # reloaded to get rid of cached array in sample
    sample = hub.read(cat_path)
    out, shape = serialize_sample_object(
        sample, compression, None, "uint16", "generic", 16 * MB
    )
    arr_deserialized = decompress_array(out, compression=compression).reshape(shape)
    np.testing.assert_array_equal(arr, arr_deserialized)

    # reloaded to get rid of cached array in sample
    sample = hub.read(cat_path)
    out, shape = serialize_sample_object(
        sample, compression, None, "uint16", "generic", 100 * KB
    )
    assert isinstance(out, SampleTiles)
    out_list = [out.yield_tile() for _ in range(out.num_tiles)]
    np_list = [
        decompress_array(b[0], compression=compression).reshape(b[1]) for b in out_list
    ]
    tile_shape, layout_shape = out.tile_shape, out.layout_shape
    out = np_list_to_sample(np_list, shape, tile_shape, layout_shape, "uint16")
    np.testing.assert_array_equal(arr, out)


def test_sample_byte_compression(cat_path, compression="lz4"):
    sample = hub.read(cat_path)
    arr = sample.array

    # reloaded to get rid of cached array in sample
    sample = hub.read(cat_path)
    dtype = "uint16"
    out, shape = serialize_sample_object(
        sample, compression, None, dtype, "generic", 16 * MB
    )
    arr_deserialized = decompress_array(out, shape, dtype, compression).reshape(shape)
    np.testing.assert_array_equal(arr, arr_deserialized)

    # reloaded to get rid of cached array in sample
    sample = hub.read(cat_path)
    out, shape = serialize_sample_object(
        sample, compression, None, dtype, "generic", 100 * KB
    )
    assert isinstance(out, SampleTiles)
    out_list = [out.yield_tile() for _ in range(out.num_tiles)]
    np_list = [decompress_array(b[0], b[1], dtype, compression) for b in out_list]
    tile_shape, layout_shape = out.tile_shape, out.layout_shape
    out = np_list_to_sample(np_list, shape, tile_shape, layout_shape, "uint16")
    np.testing.assert_array_equal(arr, out)


def test_sample_no_compression(cat_path):
    sample = hub.read(cat_path)
    arr = sample.array

    # reloaded to get rid of cached array in sample
    sample = hub.read(cat_path)
    out, shape = serialize_sample_object(
        sample, None, None, "uint16", "generic", 16 * MB
    )
    arr_deserialized = np.frombuffer(out, dtype="uint16").reshape(shape)
    np.testing.assert_array_equal(arr, arr_deserialized)

    # reloaded to get rid of cached array in sample
    sample = hub.read(cat_path)
    out, shape = serialize_sample_object(
        sample, None, None, "uint16", "generic", 100 * KB
    )
    assert isinstance(out, SampleTiles)
    out_list = [out.yield_tile() for _ in range(out.num_tiles)]
    np_list = [np.frombuffer(b[0], dtype="uint16").reshape(b[1]) for b in out_list]
    tile_shape, layout_shape = out.tile_shape, out.layout_shape
    out = np_list_to_sample(np_list, shape, tile_shape, layout_shape, "uint16")
    np.testing.assert_array_equal(arr, out)



from typing import Callable, List, Optional, Sequence, Dict
from uuid import uuid4

import hub

from hub.core.io import SampleStreaming
from hub.core.query.query import DatasetQuery
from hub.util.compute import get_compute_provider
from hub.util.dataset import map_tensor_keys
from hub.constants import QUERY_PROGRESS_UPDATE_FREQUENCY
from time import time

import inspect
import threading
from queue import Queue
from collections import defaultdict

from hub.util.exceptions import FilterError
from hub.util.hash import hash_inputs


_LAST_UPDATED_TIMES: Dict = defaultdict(time)


def _counter(id):
    """A method which returns True only every `QUERY_PROGRESS_UPDATE_FREQUENCY` seconds for each id.
    Used for sending query progress update events and writing to vds.
    """
    last_updated_time = _LAST_UPDATED_TIMES[id]
    curr_time = time()
    if curr_time - last_updated_time > QUERY_PROGRESS_UPDATE_FREQUENCY:
        _LAST_UPDATED_TIMES[id] = curr_time
        return True
    return False


def _del_counter(id):
    _LAST_UPDATED_TIMES.pop(id, None)


def _filter_function_to_query_text(filter_function):
    if isinstance(filter_function, hub.core.query.DatasetQuery):
        query_text = filter_function._query
    else:
        try:
            query_text = inspect.getsource(filter_function)
        except (OSError, TypeError):
            query_text = (
                "UDF: "
                + getattr(
                    filter_function, "__name__", filter_function.__class__.__name__
                )
                + "@"
                + str(uuid4().hex)
            )  # because function name alone is not unique enough
    return query_text


def filter_dataset(
    dataset: hub.Dataset,
    filter_function: Callable[[hub.Dataset], bool],
    num_workers: int = 0,
    scheduler: str = "threaded",
    progressbar: bool = True,
    save_result: bool = False,
    result_path: Optional[str] = None,
    result_ds_args: Optional[dict] = None,
) -> hub.Dataset:
    index_map: List[int]

    tm = time()

    query_text = _filter_function_to_query_text(filter_function)
    vds = (
        dataset._get_empty_vds(result_path, query=query_text, **(result_ds_args or {}))
        if save_result
        else None
    )

    index_map = None  # type: ignore
    try:
        if num_workers > 0:
            index_map = filter_with_compute(
                dataset,
                filter_function,
                num_workers,
                scheduler,
                progressbar,
                query_text,
                vds,
            )
        else:
            index_map = filter_inplace(
                dataset,
                filter_function,
                progressbar,
                query_text,
                vds,
            )
    except Exception as e:
        if vds:
            vds.info["error"] = str(e)
        raise e

    ds = dataset[index_map]
    ds._is_filtered_view = True

    ds._query = query_text
    ds._source_ds_idx = dataset.index.to_json()
    ds._created_at = tm
    if vds:
        ds._vds = vds
    return ds  # type: ignore [this is fine]


def _get_vds_thread(vds: hub.Dataset, queue: Queue, num_samples: int):
    """Creates a thread which writes to a vds in background.

    Args:
        vds: (hub.Dataset) The vds to write to.
        queue: (Queue) Queue to pop progress info from.
            Each item in the queue should be of form Tuple[int, bool],
            where the int is a sample index and the bool is whether
            or not to include the sample index in the vds.
        num_samples (int): Total number of samples in the source dataset.

    Returns:
        threading.Thread object
    """
    id = str(uuid4().hex)

    def loop():
        processed = 0
        while True:
            index, include = queue.get()
            vds.info["samples_processed"] += 1
            if include:
                vds.VDS_INDEX.append(index)
            processed += 1
            if processed == num_samples:
                vds.flush()
                _del_counter(id)
                break
            if _counter(id):
                vds.flush()

    return threading.Thread(target=loop)


def filter_with_compute(
    dataset: hub.Dataset,
    filter_function: Callable,
    num_workers: int,
    scheduler: str,
    progressbar: bool = True,
    query_text: Optional[str] = None,
    vds: Optional[hub.Dataset] = None,
) -> List[int]:

    blocks = SampleStreaming(dataset, tensors=map_tensor_keys(dataset)).list_blocks()
    compute = get_compute_provider(scheduler=scheduler, num_workers=num_workers)

    num_samples = len(dataset)

    if vds:
        vds.autoflush = False
        vds.info["total_samples"] = num_samples
        vds.info["samples_processed"] = 0
        vds_queue = compute.create_queue()
        vds_thread = _get_vds_thread(vds, vds_queue, num_samples)
        vds_thread.start()

    query_id = hash_inputs(dataset.path, dataset.pending_commit_id, query_text)

    progress = {"value": 0}
    # Callback for sending query progress
    def _event_callback():
        progress["value"] += 1
        if _counter(query_id):
            dataset._send_query_progress(
                query_text=query_text,
                query_id=query_id,
                progress=int(100 * progress["value"] / num_samples),
            )

    def filter_slice(indices: Sequence[int]):
        result = list()
        for i in indices:
            if filter_function(dataset[i]):
                result.append(i)
                if vds:
                    vds_queue.put((i, True))
                    _event_callback()
            elif vds:
                vds_queue.put((i, False))
                _event_callback()
        return result

    def pg_filter_slice(pg_callback, indices: Sequence[int]):
        result = list()
        for i in indices:
            if filter_function(dataset[i]):
                result.append(i)
                if vds:
                    vds_queue.put((i, True))
                    _event_callback()
            elif vds:
                vds_queue.put((i, False))
                _event_callback()
            pg_callback(1)
        return result

    result: Sequence[List[int]]
    idx: List[List[int]] = [block.indices() for block in blocks]
    if vds:
        dataset._send_query_progress(
            query_text=query_text, query_id=query_id, start=True, progress=0
        )

    try:
        if progressbar:
            result = compute.map_with_progressbar(pg_filter_slice, idx, total_length=len(dataset))  # type: ignore
        else:
            result = compute.map(filter_slice, idx)  # type: ignore
        index_map = [k for x in result for k in x]  # unfold the result map
        if vds:
            dataset._send_query_progress(
                query_text=query_text,
                query_id=query_id,
                end=True,
                progress=100,
                status="success",
            )
    except Exception as e:
        if vds:
            dataset._send_query_progress(
                query_text=query_text,
                query_id=query_id,
                end=True,
                progress=100,
                status="failed",
            )
        raise FilterError(e)

    finally:
        compute.close()
        if vds:
            if hasattr(vds_queue, "close"):
                vds_queue.close()
        _del_counter(query_id)
    if vds:
        vds.autoflush = True
        vds_thread.join()
    return index_map


def filter_inplace(
    dataset: hub.Dataset,
    filter_function: Callable,
    progressbar: bool,
    query_text: Optional[str] = None,
    vds: Optional[hub.Dataset] = None,
) -> List[int]:
    index_map: List[int] = list()

    it = enumerate(dataset)
    num_samples = len(dataset)
    if vds:
        vds.autoflush = False
        vds.info["total_samples"] = len(dataset)
        vds.info["samples_processed"] = 0
        vds_queue: Queue = Queue()
        vds_thread = _get_vds_thread(vds, vds_queue, num_samples)
        vds_thread.start()
    if progressbar:
        from tqdm import tqdm  # type: ignore

        it = tqdm(it, total=num_samples)

    query_id = hash_inputs(dataset.path, dataset.pending_commit_id, query_text)

    if vds:
        dataset._send_query_progress(
            query_text=query_text, query_id=query_id, start=True, progress=0
        )

    try:
        for i, sample_in in it:
            if filter_function(sample_in):
                index_map.append(i)
                if vds:
                    vds_queue.put((i, True))
            elif vds:
                vds_queue.put((i, False))
            if vds and _counter(query_id):
                dataset._send_query_progress(
                    query_text=query_text,
                    query_id=query_id,
                    progress=int(i * 100 / num_samples),
                    status="success",
                )
        if vds:
            dataset._send_query_progress(
                query_text=query_text,
                query_id=query_id,
                end=True,
                progress=100,
                status="success",
            )
    except Exception as e:
        if vds:
            dataset._send_query_progress(
                query_text=query_text,
                query_id=query_id,
                end=True,
                progress=100,
                status="failed",
            )
        raise (e)
    finally:
        _del_counter(query_id)

    if vds:
        vds.autoflush = True
        vds_thread.join()
    return index_map


def query_dataset(
    dataset: hub.Dataset,
    query: str,
    num_workers: int = 0,
    scheduler: str = "threaded",
    progressbar: bool = True,
    save_result: bool = False,
    result_path: Optional[str] = None,
    result_ds_args: Optional[Dict] = None,
) -> hub.Dataset:
    index_map: List[int]

    vds = (
        dataset._get_empty_vds(result_path, query=query, **(result_ds_args or {}))
        if save_result
        else None
    )
    index_map = query_inplace(dataset, query, progressbar, num_workers, scheduler, vds)
    ret = dataset[index_map]  # type: ignore [this is fine]
    ret._query = query
    if vds:
        ret._vds = vds
    return ret


def query_inplace(
    dataset: hub.Dataset,
    query: str,
    progressbar: bool,
    num_workers: int,
    scheduler: str,
    vds: Optional[hub.Dataset] = None,
) -> List[int]:

    num_samples = len(dataset)
    compute = (
        get_compute_provider(scheduler=scheduler, num_workers=num_workers)
        if num_workers > 0
        else None
    )
    query_id = hash_inputs(dataset.path, dataset.pending_commit_id, query)

    if vds:
        vds.autoflush = False
        vds.info["total_samples"] = num_samples
        vds.info["samples_processed"] = 0
        vds_queue = Queue() if num_workers == 0 else compute.create_queue()  # type: ignore
        vds_thread = _get_vds_thread(vds, vds_queue, num_samples)
        vds_thread.start()
        dataset._send_query_progress(
            query_text=query, query_id=query_id, start=True, progress=0
        )

    num_processed = {"value": 0}

    def update_vds(idx, include):
        if vds:
            vds_queue.put((idx, include))
            num_processed["value"] += 1
            if _counter(query_id):
                dataset._send_query_progress(
                    query_text=query,
                    query_id=query_id,
                    progress=int(num_processed["value"] * 100 / num_samples),
                    status="success",
                )

    class QuerySlice:
        def __init__(self, offset, size, dataset, query) -> None:
            self.offset = offset
            self.size = size
            self.dataset = dataset
            self.query = query

        def slice_dataset(self):
            return self.dataset[self.offset : (self.offset + self.size)]

    def subquery(query_slice: QuerySlice):
        dataset = query_slice.slice_dataset()
        query = query_slice.query

        if progressbar:
            from tqdm import tqdm

            bar = tqdm(total=len(dataset))

            def update(idx, include):
                bar.update(1)
                update_vds(idx, include)

            try:
                ds_query = DatasetQuery(dataset, query, update)
                ret = ds_query.execute()
            finally:
                bar.close()
        else:
            ret = DatasetQuery(dataset, query, update_vds).execute()
        return ret

    def pg_subquery(pg_callback, query_slice):
        def update(idx, include):
            update_vds(idx, include)
            pg_callback(1)

        dataset = query_slice.slice_dataset()
        ds_query = DatasetQuery(dataset, query, progress_callback=update)
        return ds_query.execute()

    try:
        if num_workers == 0:
            index_map = subquery(QuerySlice(0, len(dataset), dataset, query))
        else:
            compute = get_compute_provider(scheduler=scheduler, num_workers=num_workers)

            btch = len(dataset) // num_workers
            subdatasets = [
                QuerySlice(idx * btch, btch, dataset, query)
                for idx in range(0, num_workers)
            ]

            if progressbar:
                result = compute.map_with_progressbar(pg_subquery, subdatasets, total_length=num_samples)  # type: ignore
            else:
                result = compute.map(subquery, subdatasets)  # type: ignore

            index_map = []
            for ls in result:
                index_map.extend(ls)

    except Exception as e:
        dataset._send_query_progress(
            query_text=query,
            query_id=query_id,
            end=True,
            progress=100,
            status="failed",
        )
        raise e
    finally:
        if vds and hasattr(vds_queue, "close"):
            vds_queue.close()
        if compute:
            compute.close()
        _del_counter(query_id)
    dataset._send_query_progress(
        query_text=query,
        query_id=query_id,
        end=True,
        progress=100,
        status="success",
    )
    if vds:
        vds.autoflush = True
        vds_thread.join()
    return index_map



from hub.api.tests.test_api import MAX_FLOAT_DTYPE
import numpy as np
import hub


def test_version(local_ds_generator):
    ds = local_ds_generator()
    assert ds.meta.version == hub.__version__

    # persistence
    ds = local_ds_generator()
    assert ds.meta.version == hub.__version__


def test_subsequent_updates(local_ds_generator):
    with local_ds_generator() as ds:
        ds.create_tensor("tensor")  # dtype is not specified

    assert ds.tensor.dtype == None

    with local_ds_generator() as ds:
        ds.tensor.extend(np.ones((5, 100, 100)))

    # dtype is auto-specified
    assert ds.tensor.dtype == MAX_FLOAT_DTYPE

    ds = local_ds_generator()
    assert len(ds) == 5

    with local_ds_generator() as ds:
        for _ in range(5):
            ds.tensor.append(np.ones((100, 100)))

    ds = local_ds_generator()
    assert len(ds) == 10
    assert ds.tensor.shape == (10, 100, 100)

    ds = local_ds_generator()

    with local_ds_generator() as ds:
        for _ in range(5):
            ds.tensor.append(np.ones((100, 200)))

    assert ds.tensor.shape == (15, 100, None)
    si = ds.tensor.shape_interval
    assert si.lower == (15, 100, 100)
    assert si.upper == (15, 100, 200)



import os
import io
import pytest
import numpy as np
from PIL import Image  # type: ignore

from hub.core.tiling.optimizer import get_tile_shape


def test_tile_shape():
    x = np.random.random((1000, 500, 3))
    tile_shape = get_tile_shape(
        sample_shape=x.shape,
        sample_size=x.nbytes,
        chunk_size=x.nbytes / 16,
        exclude_axes=-1,
    )
    assert np.prod(tile_shape) * x.itemsize == x.nbytes / 16
    assert tile_shape[2] == 3


@pytest.mark.parametrize("compression", ["jpeg", "png", "tiff"])
def test_tile_shape_compressed(compression, compressed_image_paths):
    path = compressed_image_paths[compression][0]
    sample_size = os.path.getsize(path)
    arr = np.array(Image.open(path))
    sample_shape = arr.shape
    chunk_size = sample_size / 10
    tile_shape = get_tile_shape(
        sample_shape=sample_shape,
        sample_size=sample_size,
        chunk_size=chunk_size,
        exclude_axes=-1,
    )
    tile = arr[: tile_shape[0], : tile_shape[1]]
    bio = io.BytesIO()
    Image.fromarray(tile).save(bio, compression)
    bio.seek(0)
    nbytes_compressed = len(bio.read())
    assert (
        nbytes_compressed < chunk_size
    )  # tiles wont be optimal for small compressed samples.


@pytest.mark.parametrize("compression", ["jpeg", "png"])
def test_tile_shape_large_compressed(compression):
    arr = np.random.randint(0, 256, (1000, 1000, 3), dtype=np.uint8)
    bio = io.BytesIO()
    Image.fromarray(arr).save(bio, compression)
    bio.seek(0)
    sample_size = len(bio.read())
    chunk_size = sample_size / 10
    tile_shape = get_tile_shape(
        sample_shape=arr.shape,
        sample_size=sample_size,
        chunk_size=chunk_size,
        exclude_axes=-1,
    )
    tile = arr[: tile_shape[0], : tile_shape[1]]
    bio = io.BytesIO()
    Image.fromarray(tile).save(bio, compression)
    bio.seek(0)
    tile_size = len(bio.read())
    assert 1 <= chunk_size / tile_size < 2



import click

from hub import __version__
from hub.cli.auth import login, logout, register, reporting
from hub.cli.list_datasets import list_datasets


@click.group()
@click.version_option(__version__, message="%(prog)s %(version)s")
@click.pass_context
def cli(ctx):
    pass


def add_auth_commands(cli):
    cli.add_command(login)
    cli.add_command(logout)
    cli.add_command(register)
    cli.add_command(reporting)
    cli.add_command(list_datasets)


add_auth_commands(cli)



import os
import re
from setuptools import find_packages, setup

project_name = "hub"


this_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_directory, "hub/requirements/common.txt")) as f:
    requirements = f.readlines()

with open(os.path.join(this_directory, "hub/requirements/tests.txt")) as f:
    tests = f.readlines()

with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


req_map = {
    b: a
    for a, b in (
        re.findall(r"^(([^!=<>~]+)(?:[!=<>~].*)?$)", x.strip("\n"))[0]
        for x in requirements
    )
}

# Add optional dependencies to this dict without version. Version should be specified in requirements.txt
extras = {
    "audio": ["av"],
    "video": ["av"],
    "av": ["av"],
    "gcp": ["google-cloud-storage", "google-auth", "google-auth-oauthlib"],
    "dicom": ["pydicom"],
    "visualizer": ["IPython", "flask"],
    "gdrive": [
        "google-api-python-client",
        "oauth2client",
        "google-auth",
        "google-auth-oauthlib",
    ],
    "point_cloud": ["laspy"],
    "deeplake": ["deeplake"],
}

all_extras = {r for v in extras.values() for r in v}
install_requires = [req_map[r] for r in req_map if r not in all_extras]
extras_require = {k: [req_map[r] for r in v] for k, v in extras.items()}
extras_require["all"] = [req_map[r] for r in all_extras]


init_file = os.path.join(project_name, "__init__.py")


def get_property(prop):
    result = re.search(
        # find variable with name `prop` in the __init__.py file
        rf'{prop}\s*=\s*[\'"]([^\'"]*)[\'"]',
        open(init_file).read(),
    )
    return result.group(1)


config = {
    "name": project_name,
    "version": get_property("__version__"),
    "description": "Activeloop Hub",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "author": "activeloop.ai",
    "author_email": "support@activeloop.ai",
    "packages": find_packages(),
    "install_requires": install_requires,
    "extras_require": extras_require,
    "tests_require": tests,
    "include_package_data": True,
    "zip_safe": False,
    "entry_points": {"console_scripts": ["activeloop = hub.cli.commands:cli"]},
    "dependency_links": [],
    "project_urls": {
        "Documentation": "https://docs.activeloop.ai/",
        "Source": "https://github.com/activeloopai/Hub",
    },
}

setup(**config)



def max_array_length(arr_max, arr_to_compare):  # helper for __str__
    for i in range(len(arr_max)):
        str_length = len(arr_to_compare[i])
        if arr_max[i] < str_length:
            arr_max[i] = str_length
    return arr_max


def get_string(
    table_array, max_arr
):  # gets string from array of arrays as a table (helper for __str__)
    temp_str = ""
    for row in table_array:
        temp_str += "\n"
        for col_no in range(len(row)):
            max_col = max_arr[col_no]
            length = len(row[col_no])
            starting_loc = (max_col - length) // 2
            temp_str += (
                " " * starting_loc
                + row[col_no]
                + " " * (max_col - length - starting_loc)
            )
    return temp_str


def summary_tensor(tensor):
    head = [
        "htype",
        "shape",
        "dtype",
        "compression",
    ]
    divider = ["-------"] * 4
    max_column_length = [7, 7, 7, 7]

    tensor_htype = tensor.htype
    if tensor_htype == None:
        tensor_htype = "None"

    shape = tensor.shape
    tensor_shape = str(tensor.shape_interval if None in shape else shape)

    tensor_compression = tensor.meta.sample_compression
    if tensor_compression == None:
        tensor_compression = "None"

    if tensor.dtype == None:
        tensor_dtype = "None"
    else:
        tensor_dtype = tensor.dtype.name

    self_array = [
        head,
        divider,
        [
            tensor_htype,
            tensor_shape,
            tensor_dtype,
            tensor_compression,
        ],
    ]
    # adding information about tensors
    max_column_length = max_array_length(
        max_column_length, self_array[2]
    )  # 3rd element of self_array corresponds to tensor att
    max_column_length = [elem + 2 for elem in max_column_length]
    return get_string(self_array, max_column_length)


def summary_dataset(dataset):
    head = [
        "tensor",
        "htype",
        "shape",
        "dtype",
        "compression",
    ]  # To add more attributes change head, divider, row_array and table_str
    divider = ["-------"] * 5
    tensor_dict = dataset.tensors
    max_column_length = [7, 7, 7, 7, 7]  # length of "attribute name" length
    count = 0
    table_array = [head, divider]  # collects all the rows
    for tensor_name in tensor_dict:
        tensor_object = tensor_dict[tensor_name]

        tensor_htype = tensor_object.htype
        if tensor_htype == None:
            tensor_htype = "None"

        shape = tensor_object.shape
        tensor_shape = str(tensor_object.shape_interval if None in shape else shape)

        tensor_compression = tensor_object.meta.sample_compression
        if tensor_compression == None:
            tensor_compression = "None"

        if tensor_object.dtype == None:
            tensor_dtype = "None"
        else:
            tensor_dtype = tensor_object.dtype.name

        row_array = [
            tensor_name,
            tensor_htype,
            tensor_shape,
            tensor_dtype,
            tensor_compression,
        ]
        table_array.append(row_array)
        max_column_length = max_array_length(max_column_length, row_array)
        count += 1
    max_column_length = [elem + 2 for elem in max_column_length]

    return get_string(table_array, max_column_length)



from hub.constants import KB
from hub.util.exceptions import TensorInvalidSampleShapeError
import pytest
from typing import Callable
from hub.tests.common import assert_array_lists_equal
import numpy as np
import hub


def _add_dummy_mnist(ds, **kwargs):
    compression = kwargs.get(
        "compression", {"image_compression": {"sample_compression": None}}
    )
    ds.create_tensor("images", htype="image", **compression["image_compression"])
    ds.create_tensor(
        "labels", htype="class_label", **compression.get("label_compression", {})
    )

    ds.images.extend(np.ones((10, 28, 28), dtype=np.uint8))
    ds.labels.extend(np.ones(10, dtype=np.uint8))

    return ds


def _make_update_assert_equal(
    ds_generator: Callable,
    tensor_name: str,
    index,
    value,
    check_persistence: bool = True,
):
    """Updates a tensor and checks that the data is as expected.

    Example update:
        >>> ds.tensor[0:5] = [1, 2, 3, 4, 5]

    Args:
        ds_generator (Callable): Function that returns a new dataset object with each call.
        tensor_name (str): Name of the tensor to be updated.
        index (Any): Any value that can be used as an index for updating (`ds.tensor[index] = value`).
        value (Any): Any value that can be used as a value for updating (`ds.tensor[index] = value`).
        check_persistence (bool): If True, the update will be tested to make sure it can be serialized/deserialized.
    """

    ds = ds_generator()
    assert len(ds) == 10

    tensor = ds[tensor_name]
    expected = tensor.numpy(aslist=True)

    # this is necessary because `expected` uses `aslist=True` to handle dynamic cases.
    # with `aslist=False`, this wouldn't be necessary.
    expected_value = value
    if hasattr(value, "__len__"):
        if len(value) == 1:
            expected_value = value[0]

    # make updates
    tensor[index] = value
    expected[index] = expected_value

    # non-persistence check
    actual = tensor.numpy(aslist=True)
    assert_array_lists_equal(actual, expected)
    assert len(ds) == 10

    if check_persistence:
        ds = ds_generator()
        tensor = ds[tensor_name]
        actual = tensor.numpy(aslist=True)
        assert_array_lists_equal(actual, expected)

        # make sure no new values are recorded
        ds = ds_generator()
        assert len(ds) == 10


@pytest.mark.parametrize(
    "compression",
    [
        {
            "image_compression": {"sample_compression": None},
        },
        {
            "image_compression": {"sample_compression": None},
            "label_compression": {"sample_compression": "lz4"},
        },
        {
            "image_compression": {"sample_compression": None},
            "label_compression": {"chunk_compression": "lz4"},
        },
        {"image_compression": {"sample_compression": "png"}},
        {"image_compression": {"chunk_compression": "png"}},
        {"image_compression": {"sample_compression": "lz4"}},
        {"image_compression": {"chunk_compression": "lz4"}},
    ],
)
def test(local_ds_generator, compression):
    gen = local_ds_generator

    _add_dummy_mnist(gen(), **compression)

    # update single sample
    _make_update_assert_equal(
        gen, "images", -1, np.ones((28, 28), dtype="uint8") * 75
    )  # same shape
    _make_update_assert_equal(
        gen, "images", -1, np.ones((28, 28), dtype="uint8") * 75
    )  # same shape
    _make_update_assert_equal(
        gen, "images", 0, np.ones((28, 25), dtype="uint8") * 5
    )  # new shape
    _make_update_assert_equal(
        gen, "images", 0, np.ones((32, 32), dtype="uint8") * 5
    )  # new shape
    _make_update_assert_equal(
        gen, "images", -1, np.ones((0, 0), dtype="uint8")
    )  # empty sample (new shape)
    _make_update_assert_equal(gen, "labels", -5, np.uint8(99))
    _make_update_assert_equal(gen, "labels", 0, np.uint8(5))

    # update a range of samples
    x = np.arange(3 * 28 * 28).reshape((3, 28, 28)).astype("uint8")
    _make_update_assert_equal(gen, "images", slice(0, 3), x)  # same shapes
    _make_update_assert_equal(
        gen, "images", slice(3, 5), np.zeros((2, 5, 28), dtype="uint8")
    )  # new shapes
    _make_update_assert_equal(
        gen, "images", slice(3, 5), np.zeros((2, 5, 28), dtype=int).tolist()
    )  # test downcasting python scalars
    _make_update_assert_equal(
        gen, "images", slice(3, 5), np.zeros((2, 5, 28), dtype=np.ubyte).tolist()
    )  # test upcasting
    _make_update_assert_equal(
        gen, "images", slice(3, 5), np.zeros((2, 0, 0), dtype="uint8")
    )  # empty samples (new shape)
    _make_update_assert_equal(gen, "labels", slice(0, 5), [1, 2, 3, 4, 5])

    # update a range of samples with dynamic samples
    _make_update_assert_equal(
        gen,
        "images",
        slice(7, 10),
        [
            np.ones((28, 50), dtype="uint8") * 5,
            np.ones((0, 5), dtype="uint8"),
            np.ones((1, 1), dtype="uint8") * 10,
        ],
    )

    ds = gen()
    assert ds.images.shape_interval.lower == (10, 0, 0)
    assert ds.images.shape_interval.upper == (10, 32, 50)


@pytest.mark.parametrize("images_compression", [None, "png"])
def test_hub_read(local_ds_generator, images_compression, cat_path, flower_path):
    gen = local_ds_generator

    ds = gen()
    ds.create_tensor("images", htype="image", sample_compression=images_compression)
    ds.images.extend(np.zeros((10, 0, 0, 0), dtype=np.uint8))

    ds.images[0] = hub.read(cat_path)
    np.testing.assert_array_equal(ds.images[0].numpy(), hub.read(cat_path).array)

    ds.images[1] = hub.read(flower_path)
    np.testing.assert_array_equal(ds.images[1].numpy(), hub.read(flower_path).array)

    ds.images[8:10] = [hub.read(cat_path), hub.read(flower_path)]
    assert_array_lists_equal(
        ds.images[8:10].numpy(aslist=True),
        [hub.read(cat_path).array, hub.read(flower_path).array],
    )

    assert ds.images.shape_interval.lower == (10, 0, 0, 0)
    assert ds.images.shape_interval.upper == (10, 900, 900, 4)

    assert len(ds.images) == 10


def test_pre_indexed_tensor(memory_ds):
    """A pre-indexed tensor update means the tensor was already indexed into, and an update is being made to that tensor view."""

    tensor = memory_ds.create_tensor("tensor")

    tensor.append([0, 1, 2])
    tensor.append([3, 4, 5, 6, 7])
    tensor.append([8, 5])
    tensor.append([9, 10, 11])
    tensor.append([12, 13, 14, 15, 16])
    tensor.append([17, 18, 19, 20, 21])

    tensor[0:5][0] = [99, 98, 97]
    tensor[5:10][0] = [44, 44, 44, 44]
    tensor[4:10][0:2] = [[44, 44, 44, 44], [33]]

    np.testing.assert_array_equal([99, 98, 97], tensor[0])
    np.testing.assert_array_equal([44, 44, 44, 44], tensor[4])
    np.testing.assert_array_equal([33], tensor[5])

    assert tensor.shape_interval.lower == (6, 1)
    assert tensor.shape_interval.upper == (6, 5)
    assert len(tensor) == 6


def test_failures(memory_ds):
    _add_dummy_mnist(memory_ds)

    # primary axis doesn't match
    with pytest.raises(ValueError):
        memory_ds.images[0:3] = np.zeros((25, 30), dtype="uint8")
    with pytest.raises(ValueError):
        memory_ds.images[0:3] = np.zeros((2, 25, 30), dtype="uint8")
    with pytest.raises(TensorInvalidSampleShapeError):
        memory_ds.images[0] = np.zeros((2, 25, 30), dtype="uint8")
    with pytest.raises(ValueError):
        memory_ds.labels[0:3] = [1, 2, 3, 4]

    # dimensionality doesn't match
    with pytest.raises(TensorInvalidSampleShapeError):
        memory_ds.images[0:5] = np.zeros((5, 30), dtype="uint8")
    with pytest.raises(TensorInvalidSampleShapeError):
        memory_ds.labels[0:5] = np.zeros((5, 2, 3), dtype="uint8")

    # make sure no data changed
    assert len(memory_ds.images) == 10
    assert len(memory_ds.labels) == 10
    np.testing.assert_array_equal(
        memory_ds.images.numpy(), np.ones((10, 28, 28), dtype="uint8")
    )
    np.testing.assert_array_equal(
        memory_ds.labels.numpy(), np.ones((10, 1), dtype="uint8")
    )
    assert memory_ds.images.shape == (10, 28, 28)
    assert memory_ds.labels.shape == (10, 1)


def test_warnings(memory_ds):
    tensor = memory_ds.create_tensor(
        "tensor", max_chunk_size=8 * KB, tiling_threshold=4 * KB
    )

    tensor.extend(np.ones((10, 12, 12), dtype="int32"))

    # this update makes (small) suboptimal chunks
    with pytest.warns(UserWarning):
        tensor[0:5] = np.zeros((5, 0, 0), dtype="int32")

    # this update makes (large) suboptimal chunks
    with pytest.warns(UserWarning):
        tensor[:] = np.zeros((10, 32, 31), dtype="int32")


@pytest.mark.parametrize(
    "compression",
    [
        {"sample_compression": None},
        {"sample_compression": "png"},
        {"chunk_compression": "png"},
        {"sample_compression": "lz4"},
        {"chunk_compression": "lz4"},
    ],
)
def test_inplace_updates(memory_ds, compression):
    ds = memory_ds
    ds.create_tensor("x", **compression)
    ds.x.extend(np.zeros((5, 32, 32, 3), dtype="uint8"))
    ds.x += 1
    np.testing.assert_array_equal(ds.x.numpy(), np.ones((5, 32, 32, 3)))
    ds.x += ds.x
    np.testing.assert_array_equal(ds.x.numpy(), np.ones((5, 32, 32, 3)) * 2)
    ds.x *= np.zeros(3, dtype="uint8")
    np.testing.assert_array_equal(ds.x.numpy(), np.zeros((5, 32, 32, 3)))
    ds.x += 6
    ds.x //= 2
    np.testing.assert_array_equal(ds.x.numpy(), np.ones((5, 32, 32, 3)) * 3)
    ds.x[:3] *= 0
    np.testing.assert_array_equal(
        ds.x.numpy(),
        np.concatenate([np.zeros((3, 32, 32, 3)), np.ones((2, 32, 32, 3)) * 3]),
    )

    # Different shape
    ds.x.append(np.zeros((100, 50, 3), dtype="uint8"))
    ds.x[5] += 1
    np.testing.assert_array_equal(ds.x[5].numpy(), np.ones((100, 50, 3)))
    np.testing.assert_array_equal(
        ds.x[:5].numpy(),
        np.concatenate([np.zeros((3, 32, 32, 3)), np.ones((2, 32, 32, 3)) * 3]),
    )
    ds.x[:5] *= 0
    np.testing.assert_array_equal(ds.x[:5].numpy(), np.zeros((5, 32, 32, 3)))
    np.testing.assert_array_equal(ds.x[5].numpy(), np.ones((100, 50, 3)))


@pytest.mark.parametrize("aslist", (True, False))
@pytest.mark.parametrize(
    "idx", [3, slice(None), slice(5, 9), slice(3, 7, 2), [3, 7, 6, 4]]
)
def test_sequence_htype(memory_ds, aslist, idx):
    ds = memory_ds
    with ds:
        ds.create_tensor("x", htype="sequence")
        for _ in range(10):
            ds.x.append([np.ones((3, 7)) for _ in range(5)])
    assert ds.x[0].numpy().shape == (5, 3, 7)
    ds.x[idx] += 1
    expected = np.ones((10, 5, 3, 7))
    expected[idx] += 1
    np.testing.assert_array_equal(np.array(ds.x.numpy(aslist=aslist)), expected)
    assert ds.x.shape == (10, 5, 3, 7)


def test_sequence_htype_with_broadcasting(memory_ds):
    ds = memory_ds
    with ds:
        arr = np.random.randint(0, 10, (2, 7, 9))
        expected = arr.reshape(1, *arr.shape).repeat(15, 0).reshape(5, 3, *arr.shape)
        ds.create_tensor("x", htype="sequence")
        for _ in range(5):
            ds.x.append([arr] * 3)
        assert ds.x.shape == expected.shape

        def _check():
            np.testing.assert_array_equal(ds.x.numpy(), expected)

        ds.x += 3
        expected += 3
        _check()
        ds.x[:3] *= 2
        expected[:3] *= 2
        _check()
        ds.x[0][2][1] *= 7
        expected[0][2][1] *= 7
        _check()
        ds.x[1][:2][0] = np.ones((2, 7, 9), np.int32) * 13
        expected[1][:2][0] = np.ones((2, 7, 9), np.int32) * 13
        _check()
        expected[:, :] = np.ones((2, 7, 9), np.int32) * 17
        ds.x[:, :] = np.ones((2, 7, 9), np.int32) * 17
        _check()
        expected[:, 1:] = np.ones((2, 2, 7, 9), np.int32) - 9
        ds.x[:, 1:] = np.ones((2, 2, 7, 9), np.int32) - 9
        _check()
        expected[:] = np.zeros_like(expected) * 13
        ds.x[:] = expected
        _check()
        ds.x[:] = expected.reshape(1, 1, 1, 1, *expected.shape)
        _check()


@pytest.mark.parametrize("shape", [(13, 17, 3), (1007, 3001, 3)])
def test_sequence_htype_with_hub_read(local_ds, shape, compressed_image_paths):
    ds = local_ds
    imgs = list(map(hub.read, compressed_image_paths["jpeg"][:3]))
    new_imgs = list(map(hub.read, compressed_image_paths["jpeg"][3:6]))
    arrs = np.random.randint(0, 256, (5, *shape), dtype=np.uint8)
    with ds:
        ds.create_tensor("x", htype="sequence[image]", sample_compression="png")
        for i in range(5):
            if i % 2:
                ds.x.append(imgs)
            else:
                ds.x.append(arrs)
    ds.x[0][1] = new_imgs[1]
    np.testing.assert_array_equal(ds.x[0][1].numpy(), new_imgs[1].array)
    ds.x[1] = new_imgs
    for t, img in zip(ds.x[1], new_imgs):
        np.testing.assert_array_equal(t.numpy(), img.array)


def test_byte_positions_encoder_update_bug(memory_ds):
    ds = memory_ds
    with ds:
        ds.create_tensor("abc")
        for i in range(11):
            ds.abc.append(np.ones((1, 1)))
        ds.abc[10] = np.ones((5, 5))
        ds.abc[0] = np.ones((2, 2))
    assert ds.abc[10].numpy().shape == (5, 5)
    assert ds.abc[0].numpy().shape == (2, 2)
    for i in range(1, 10):
        assert ds.abc[i].numpy().shape == (1, 1)


@pytest.mark.parametrize(
    "args",
    [
        {},
        {"sample_compression": "lz4"},
        {"chunk_compression": "lz4"},
        {"sample_compression": "png"},
        {"chunk_compression": "png"},
    ],
)
@pytest.mark.parametrize("htype", ["generic", "sequence"])
def test_update_partial(memory_ds, htype, args):
    ds = memory_ds
    with ds:
        ds.create_tensor("x", htype=htype, **args)
        ds.x.append(np.ones((10, 10, 3), dtype=np.uint8))
        ds.x[0][0:2, 0:3, :1] = np.zeros((2, 3, 1), dtype=np.uint8)
    assert ds.x[0].shape == (10, 10, 3)
    arr = ds.x[0].numpy()
    exp = np.ones((10, 10, 3), dtype=np.uint8)
    exp[0:2, 0:3, 0] *= 0
    np.testing.assert_array_equal(arr, exp)
    with ds:
        ds.x[0][1] += 1
        ds.x[0][1] *= 3
    exp[1] += 1
    exp[1] *= 3
    arr = ds.x[0].numpy()
    np.testing.assert_array_equal(arr, exp)



from abc import ABC, abstractmethod
from hub.util.path import find_root
from pathlib import Path
import os


class StructuredDataset(ABC):
    """Initializes a structured dataset.

    Args:
        source (str): The local path to folder or file containing a structured dataset and of the form ./path/to/dataset or ~/path/to/dataset or path/to/dataset.
    """

    def __init__(self, source):
        if isinstance(source, str):
            if os.path.isdir(source):
                self.source = Path(find_root(source))
            else:
                self.source = Path(source)
        else:
            self.source = source



import sys


def is_notebook():
    """Whether running in a notebook."""
    try:
        from IPython import get_ipython  # type: ignore

        if get_ipython() is None:
            return False
    except ImportError:
        return False
    return True


def is_jupyter():
    """Whether running in a Jupyter notebook."""
    if not is_notebook():
        return False
    from IPython import get_ipython

    if "terminal" in get_ipython().__module__ or "spyder" in sys.modules:
        return False
    return True


def is_colab():
    """Whether running in a colab notebook."""
    return "google.colab" in sys.modules


def video_html(src, alt):
    import IPython  # type: ignore

    html = f"""<video alt="{alt}" width=500 controls autoplay seek loop>
                    <source src="{src}" type="video/mp4">
                </video>
            """
    return IPython.display.HTML(html)



import hub
from hub.core.chunk.base_chunk import BaseChunk
from hub.core.chunk_engine import ChunkEngine
from hub.core.compression import _read_video_shape, _decompress_video
from hub.core.index.index import Index
from hub.core.link_creds import LinkCreds
from hub.core.linked_sample import LinkedSample
from hub.core.meta.encode.creds import CredsEncoder
from hub.core.storage import LRUCache
from hub.core.tensor_link import read_linked_sample
from hub.util.exceptions import ReadOnlyModeError, UnableToReadFromUrlError
from hub.util.keys import get_creds_encoder_key
from hub.util.link import get_path_creds_key, save_link_creds
from hub.util.video import normalize_index
import numpy as np
from typing import Optional, Dict, Any, Tuple


def retry_refresh_managed_creds(fn):
    def wrapper(chunk_engine, global_sample_index, *args, **kwargs):
        try:
            return fn(chunk_engine, global_sample_index, *args, **kwargs)
        except UnableToReadFromUrlError:
            sample_creds_encoded = chunk_engine.creds_encoder.get_encoded_creds_key(
                global_sample_index
            )

            link_creds: LinkCreds = chunk_engine.link_creds
            sample_creds_key = link_creds.get_creds_key(sample_creds_encoded)

            if sample_creds_key in link_creds.managed_creds_keys:
                link_creds.refresh_managed_creds(sample_creds_key)
            else:
                raise
            return fn(chunk_engine, global_sample_index, *args, **kwargs)

    return wrapper


class LinkedChunkEngine(ChunkEngine):
    def __init__(
        self,
        key: str,
        cache: LRUCache,
        version_state: Dict[str, Any],
        link_creds: LinkCreds,
        meta_cache: LRUCache = None,
    ):
        super().__init__(key, cache, version_state, meta_cache)
        self.link_creds = link_creds
        self._creds_encoder: Optional[CredsEncoder] = None
        self._creds_encoder_commit_id: Optional[str] = None

    @property
    def creds_encoder(self) -> CredsEncoder:
        commit_id = self.commit_id
        if self._creds_encoder is None or self._creds_encoder_commit_id != commit_id:
            commit_id = self.commit_id
            key = get_creds_encoder_key(self.key, commit_id)
            if not self.creds_encoder_exists:
                enc = CredsEncoder()
                try:
                    self.meta_cache[key] = enc
                except ReadOnlyModeError:
                    pass
            else:
                enc = self.meta_cache.get_hub_object(key, CredsEncoder)
            self._creds_encoder = enc
            self._creds_encoder_commit_id = commit_id
            self.meta_cache.register_hub_object(key, enc)
        return self._creds_encoder

    @property
    def creds_encoder_exists(self):
        commit_id = self.commit_id
        if (
            self._creds_encoder is not None
            and self._creds_encoder_commit_id == commit_id
        ):
            return True
        try:
            key = get_creds_encoder_key(self.key, commit_id)
            self.meta_cache[key]
            return True
        except KeyError:
            return False

    @property
    def is_data_cachable(self):
        return False

    def linked_sample(self, global_sample_index: int) -> LinkedSample:
        creds_encoder = self.creds_encoder
        sample_path = self.get_path(global_sample_index)
        sample_creds_encoded = creds_encoder.get_encoded_creds_key(global_sample_index)
        sample_creds_key = self.link_creds.get_creds_key(sample_creds_encoded)
        return LinkedSample(sample_path, sample_creds_key)

    def get_video_url(self, global_sample_index):
        creds_encoder = self.creds_encoder
        sample_path = self.get_path(global_sample_index)
        sample_creds_encoded = creds_encoder.get_encoded_creds_key(global_sample_index)
        sample_creds_key = self.link_creds.get_creds_key(sample_creds_encoded)
        storage = None
        if sample_path.startswith(("gcs://", "gcp://", "s3://")):
            provider_type = "s3" if sample_path.startswith("s3://") else "gcs"
            storage = self.link_creds.get_storage_provider(
                sample_creds_key, provider_type
            )
            url = storage.get_presigned_url(sample_path, full=True)
        else:
            url = sample_path
        return url

    def get_video_sample(self, global_sample_index, index):
        url = self.get_video_url(global_sample_index)
        squeeze = isinstance(index, int)
        shape = _read_video_shape(url)
        sub_index = index.values[1].value if len(index.values) > 1 else None  # type: ignore
        start, stop, step, reverse = normalize_index(sub_index, shape[0])
        video_sample = _decompress_video(
            url,
            start,
            stop,
            step,
            reverse,
        )
        if squeeze:
            video_sample.squeeze(0)
        return video_sample

    @retry_refresh_managed_creds
    def get_basic_sample(self, global_sample_index, index, fetch_chunks=False):
        sample = self.get_hub_read_sample(global_sample_index, fetch_chunks)
        if sample is None:
            return np.ones((0,))
        return sample.array[tuple(entry.value for entry in index.values[1:])]

    def get_path(self, global_sample_index, fetch_chunks=False) -> str:
        return super().get_basic_sample(
            global_sample_index, Index(global_sample_index), fetch_chunks
        )[0]

    def get_hub_read_sample(self, global_sample_index, fetch_chunks=False):
        creds_encoder = self.creds_encoder
        sample_path = self.get_path(global_sample_index, fetch_chunks)
        if not sample_path:
            return None
        sample_creds_encoded = creds_encoder.get_encoded_creds_key(global_sample_index)
        sample_creds_key = self.link_creds.get_creds_key(sample_creds_encoded)
        return read_linked_sample(sample_path, sample_creds_key, self.link_creds, False)

    @property
    def verify(self):
        return self.tensor_meta.is_link and self.tensor_meta.verify

    def check_each_sample(self, samples, verify_creds_key_exists=True):
        link_creds = self.link_creds
        verified_samples = []
        for i, sample in enumerate(samples):
            if isinstance(sample, hub.core.tensor.Tensor) and sample.is_link:
                sample = sample._linked_sample()
                samples[i] = sample
            elif not isinstance(sample, LinkedSample) and sample is not None:
                raise TypeError(
                    f"Expected LinkedSample, got {type(sample)} instead. Use hub.link() to link samples."
                )

            path, creds_key = get_path_creds_key(sample)

            # verifies existence of creds_key
            if verify_creds_key_exists:
                link_creds.get_encoding(creds_key, path)

            if sample is None or sample.path == "":
                verified_samples.append(sample)
            else:
                verified_samples.append(
                    read_linked_sample(
                        sample.path,
                        sample.creds_key,
                        self.link_creds,
                        verify=self.verify,
                    )
                )
        return verified_samples

    def register_new_creds(self, num_samples_added, samples):
        assert isinstance(num_samples_added, int)
        link_creds = self.link_creds
        creds_encoder = self.creds_encoder
        for i in range(num_samples_added):
            sample = samples[i]
            path, creds_key = get_path_creds_key(sample)
            encoded_creds_key = link_creds.get_encoding(creds_key, path)
            creds_encoder.register_samples((encoded_creds_key,), 1)
            if link_creds.add_to_used_creds(creds_key):
                save_link_creds(self.link_creds, self.cache)
                self.link_creds.warn_missing_managed_creds()

    def update_creds(self, sample_index: int, sample: Optional[LinkedSample]):
        link_creds = self.link_creds
        path, creds_key = get_path_creds_key(sample)
        encoded_creds_key = link_creds.get_encoding(creds_key, path)
        self.creds_encoder[sample_index] = (encoded_creds_key,)
        if link_creds.add_to_used_creds(creds_key):
            save_link_creds(self.link_creds, self.cache)
            self.link_creds.warn_missing_managed_creds()

    @retry_refresh_managed_creds
    def read_shape_for_sample(self, global_sample_index: int) -> Tuple[int, ...]:
        sample = self.get_hub_read_sample(global_sample_index)
        if sample is None:
            return (0,)
        return sample.shape

    def read_sample_from_chunk(
        self,
        global_sample_index: int,
        chunk: BaseChunk,
        cast: bool = True,
        copy: bool = False,
        decompress: bool = True,
    ) -> np.ndarray:
        enc = self.chunk_id_encoder
        local_sample_index = enc.translate_index_relative_to_chunks(global_sample_index)
        sample_path = chunk.read_sample(
            local_sample_index, cast=cast, copy=copy, decompress=decompress
        )[0]

        creds_encoder = self.creds_encoder
        if not sample_path:
            return self.get_empty_sample()
        sample_creds_encoded = creds_encoder.get_encoded_creds_key(global_sample_index)
        sample_creds_key = self.link_creds.get_creds_key(sample_creds_encoded)
        return read_linked_sample(sample_path, sample_creds_key, self.link_creds, False)

    def check_link_ready(self):
        missing_keys = self.link_creds.missing_keys
        creds_used = self.link_creds.used_creds_keys
        missing_used_keys = {key for key in missing_keys if key in creds_used}
        if missing_used_keys:
            raise ValueError(
                f"Creds keys {missing_used_keys} are used in the data but not populated. Please populate the dataset using ds.populate_creds()."
            )

    def pop_item(self, index):
        self.creds_encoder.pop(index)
        return super().pop_item(index)

    def get_empty_sample(self):
        return np.ones((0,))

    @retry_refresh_managed_creds
    def read_bytes_for_sample(self, global_sample_index: int) -> bytes:
        sample = self.get_hub_read_sample(global_sample_index)
        return sample.buffer



import os
import numpy as np


BYTE_PADDING = b"\0"

# number of bytes per unit
B = 1
KB = 1000 * B
MB = 1000 * KB
GB = 1000 * MB

DEFAULT_HTYPE = "generic"

# used for requiring the user to specify a value for htype properties. notates that the htype property has no default.
REQUIRE_USER_SPECIFICATION = "require_user_specification"

# used for `REQUIRE_USER_SPECIFICATION` enforcement. this should be used instead of `None` for default user method arguments.
UNSPECIFIED = "unspecified"

SUPPORTED_MODES = ["r", "a"]

# used to show chunk size
RANDOM_CHUNK_SIZE = 8 * MB
# used to show variation between normal chunk size and maximum chunk size
RANDOM_MINIMAL_CHUNK_SIZE = 2 * MB
# used to show maximum chunk size allowed to have during random update operation
RANDOM_MAX_ALLOWED_CHUNK_SIZE = RANDOM_CHUNK_SIZE + RANDOM_MINIMAL_CHUNK_SIZE

# min chunk size is always half of `DEFAULT_MAX_CHUNK_SIZE`
DEFAULT_MAX_CHUNK_SIZE = 16 * MB

DEFAULT_TILING_THRESHOLD = 16 * MB  # Note: set to -1 to disable tiling

MIN_FIRST_CACHE_SIZE = 32 * MB
MIN_SECOND_CACHE_SIZE = 160 * MB

# without MB multiplication, meant for the dataset API that takes cache size in MBs
DEFAULT_MEMORY_CACHE_SIZE = 256
DEFAULT_LOCAL_CACHE_SIZE = 0

# maximum allowable size before `large_ok` must be passed to dataset delete methods
DELETE_SAFETY_SIZE = 1 * GB

# meta is hub-defined information, necessary for hub Datasets/Tensors to function
DATASET_META_FILENAME = "dataset_meta.json"
TENSOR_META_FILENAME = "tensor_meta.json"

# info is user-defined information, entirely optional. may be used by the visualizer
DATASET_INFO_FILENAME = "dataset_info.json"
TENSOR_INFO_FILENAME = "tensor_info.json"

DATASET_LOCK_FILENAME = "dataset_lock.lock"
DATASET_DIFF_FILENAME = "dataset_diff"
TENSOR_COMMIT_CHUNK_SET_FILENAME = "chunk_set"
TENSOR_COMMIT_DIFF_FILENAME = "commit_diff"
TIMESTAMP_FILENAME = "local_download_timestamp"


DATASET_LOCK_UPDATE_INTERVAL = 120  # seconds
DATASET_LOCK_VALIDITY = 300  # seconds

META_ENCODING = "utf8"

CHUNKS_FOLDER = "chunks"
ENCODED_TILE_NAMES_FOLDER = "tiles_index"
ENCODED_CREDS_FOLDER = "creds_index"
ENCODED_CHUNK_NAMES_FOLDER = "chunks_index"
ENCODED_SEQUENCE_NAMES_FOLDER = "sequence_index"
# unsharded naming will help with backwards compatibility
UNSHARDED_ENCODER_FILENAME = "unsharded"

ENCODING_DTYPE = np.uint32

# environment variables
ENV_HUB_DEV_USERNAME = "ACTIVELOOP_HUB_USERNAME"
ENV_HUB_DEV_PASSWORD = "ACTIVELOOP_HUB_PASSWORD"
ENV_KAGGLE_USERNAME = "KAGGLE_USERNAME"
ENV_KAGGLE_KEY = "KAGGLE_KEY"
ENV_GOOGLE_APPLICATION_CREDENTIALS = "GOOGLE_APPLICATION_CREDENTIALS"
ENV_GDRIVE_CLIENT_ID = "GDRIVE_CLIENT_ID"
ENV_GDRIVE_CLIENT_SECRET = "GDRIVE_CLIENT_SECRET"
ENV_GDRIVE_REFRESH_TOKEN = "GDRIVE_REFRESH_TOKEN"

HUB_CLOUD_DEV_USERNAME = os.getenv(ENV_HUB_DEV_USERNAME)  # type: ignore
HUB_CLOUD_DEV_PASSWORD = os.getenv(ENV_HUB_DEV_PASSWORD)

# dataset base roots for pytests
PYTEST_MEMORY_PROVIDER_BASE_ROOT = "mem://hub_pytest"
PYTEST_LOCAL_PROVIDER_BASE_ROOT = "/tmp/hub_pytest/"  # TODO: may fail for windows
PYTEST_S3_PROVIDER_BASE_ROOT = "s3://hub-2.0-tests/"
PYTEST_GCS_PROVIDER_BASE_ROOT = "gcs://snark-test/"
PYTEST_GDRIVE_PROVIDER_BASE_ROOT = (
    "gdrive://hubtest"  # TODO: personal folder, replace with hub's
)
PYTEST_HUB_CLOUD_PROVIDER_BASE_ROOT = (
    None if HUB_CLOUD_DEV_USERNAME is None else f"hub://{HUB_CLOUD_DEV_USERNAME}/"
)

# pytest options
MEMORY_OPT = "--memory-skip"
LOCAL_OPT = "--local"
S3_OPT = "--s3"
GCS_OPT = "--gcs"
GDRIVE_OPT = "--gdrive"
HUB_CLOUD_OPT = "--hub-cloud"
S3_PATH_OPT = "--s3-path"
GDRIVE_PATH_OPT = "--gdrive-path"
KEEP_STORAGE_OPT = "--keep-storage"
KAGGLE_OPT = "--kaggle"


EMERGENCY_STORAGE_PATH = "/tmp/emergency_storage"
LOCAL_CACHE_PREFIX = "~/.activeloop/cache"

# used to identify the first commit so its data will not be in similar directory structure to the rest
FIRST_COMMIT_ID = "firstdbf9474d461a19e9333c2fd19b46115348f"
VERSION_CONTROL_INFO_FILENAME_OLD = "version_control_info"
VERSION_CONTROL_INFO_FILENAME = "version_control_info.json"
VERSION_CONTROL_INFO_LOCK_FILENAME = "version_control_info.lock"

LINKED_CREDS_FILENAME = "linked_creds.json"
LINKED_CREDS_LOCK_FILENAME = "linked_creds.lock"


# when cache is full upto this threshold, it will start suggesting new indexes intelligently based on existing contents
INTELLIGENT_SHUFFLING_THRESHOLD = 0.8

TRANSFORM_PROGRESSBAR_UPDATE_INTERVAL = 5  # seconds


# If True, and if the rest of the dataset is in color (3D), then reshape a grayscale image by appending a 1 to its shape.
CONVERT_GRAYSCALE = True

PARTIAL_NUM_SAMPLES = 0.5

QUERIES_FILENAME = "queries.json"
QUERIES_LOCK_FILENAME = "queries.lock"

ALL_CLOUD_PREFIXES = ("s3://", "gcs://", "gcp://", "gs://", "gdrive://")

_ENABLE_HUB_SUB_DATASETS = False
_ENABLE_RANDOM_ASSIGNMENT = True

# Frequency for sending progress events and writing to vds
QUERY_PROGRESS_UPDATE_FREQUENCY = 5  # seconds

PYTORCH_DATALOADER_TIMEOUT = 30  # seconds

_NO_LINK_UPDATE = "___!@#_no_link_update_###"

SAMPLE_INFO_TENSOR_MAX_CHUNK_SIZE = 4 * MB

DEFAULT_READONLY = (
    os.environ.get("HUB_DEFAULT_READONLY", "false").strip().lower() == "true"
)

_UNLINK_VIDEOS = False

WANDB_INTEGRATION_ENABLED = True
WANDB_JSON_FILENMAE = "wandb.json"



import json
from typing import Any, Optional
from hub.core.meta.encode.shape import ShapeEncoder
from hub.core.serialize import (
    deserialize_sequence_or_creds_encoder,
    serialize_sequence_or_creds_encoder,
)
from hub.core.storage.hub_memory_object import HubMemoryObject


class CredsEncoder(ShapeEncoder, HubMemoryObject):
    def __init__(self):
        self.is_dirty = False
        super().__init__()

    def register_samples(self, item: Any, num_samples: int, row: Optional[int] = None):
        self.is_dirty = True
        return super().register_samples(item, num_samples)

    def __setitem__(self, local_sample_index: int, item: Any):
        self.is_dirty = True
        return super().__setitem__(local_sample_index, item)

    def get_encoded_creds_key(self, local_sample_index: int):
        return self[local_sample_index][0]

    @classmethod
    def frombuffer(cls, buffer: bytes):
        instance = cls()
        if not buffer:
            return instance
        version, ids = deserialize_sequence_or_creds_encoder(buffer, "creds")
        if ids.nbytes:
            instance._encoded = ids
        instance.version = version
        instance.is_dirty = False
        return instance

    def tobytes(self) -> memoryview:
        return memoryview(
            serialize_sequence_or_creds_encoder(self.version, self._encoded)
        )

    @property
    def nbytes(self):
        return len(self.tobytes())

    def pop(self, index):
        self.is_dirty = True
        super().pop(index)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import logging
import sys

logger = logging.getLogger("hub")


def configure_logger(debug=0):
    log_level = logging.DEBUG if debug == 1 else logging.INFO
    logger.setLevel(log_level)

    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setLevel(log_level)

    formatter = logging.Formatter("%(message)s")

    stream_handler.setFormatter(formatter)

    logger.addHandler(stream_handler)
    logger.propagate = False


configure_logger(0)



import json
import os
from pathlib import Path
from platform import machine
from typing import Any, Dict, Optional
import uuid

from hub.client.config import REPORTING_CONFIG_FILE_PATH
from hub.client.client import HubBackendClient
from hub.client.utils import get_user_name
from hub.util.bugout_token import BUGOUT_TOKEN
from humbug.consent import HumbugConsent
from humbug.report import HumbugReporter


def save_reporting_config(
    consent: bool, client_id: Optional[str] = None, username: Optional[str] = None
) -> Dict[str, Any]:
    """Modify reporting config.

    Args:
        consent (bool): Enabling and disabling sending crashes and system report to Activeloop Hub.
        client_id (str, optional): Unique client id.
        username (str, optional): Activeloop username.

    Returns:
        The configuration that it just saved.
    """
    reporting_config = {}

    if os.path.isfile(REPORTING_CONFIG_FILE_PATH):
        try:
            with open(REPORTING_CONFIG_FILE_PATH, "r") as ifp:
                reporting_config = json.load(ifp)
        except Exception:
            pass
    else:
        # We should not expect that the parent directory for the reporting configuration will exist.
        # If it doesn't exist, we create the directory, if possible.
        # This mirrors the code for the `write_token` method in hub/client/utils.py.
        path = Path(REPORTING_CONFIG_FILE_PATH)
        os.makedirs(path.parent, exist_ok=True)

    if client_id is not None and reporting_config.get("client_id") is None:
        reporting_config["client_id"] = client_id

    if reporting_config.get("client_id") is None:
        reporting_config["client_id"] = str(uuid.uuid4())

    if username is not None:
        reporting_config["username"] = username

    reporting_config["consent"] = consent

    try:
        with open(REPORTING_CONFIG_FILE_PATH, "w") as ofp:
            json.dump(reporting_config, ofp)
    except Exception:
        pass

    return reporting_config


def get_reporting_config() -> Dict[str, Any]:
    """Get an existing reporting config"""
    reporting_config: Dict[str, Any] = {"consent": False}
    try:
        if not os.path.exists(REPORTING_CONFIG_FILE_PATH):
            client_id = str(uuid.uuid4())
            reporting_config["client_id"] = client_id
            reporting_config = save_reporting_config(True, client_id)
        else:
            with open(REPORTING_CONFIG_FILE_PATH, "r") as ifp:
                reporting_config = json.load(ifp)

        # The following changes do NOT mutate the reporting_config.json file on the file system, but
        # they provide a means to report the username as the client_id (if the username is available)
        # while tracking the existing client_id as a machine_id.
        reporting_config["machine_id"] = reporting_config["client_id"]

        if (
            reporting_config.get("username") is not None
            and reporting_config["client_id"] != reporting_config["username"]
        ):
            reporting_config["client_id"] = reporting_config["username"]

    except Exception:
        # Not being able to load reporting consent should not get in the user's way. We will just
        # return the default reporting_config object in which consent is set to False.
        pass
    return reporting_config


def consent_from_reporting_config_file() -> bool:
    """Get consent settings from the existing reporting config"""
    reporting_config = get_reporting_config()
    return reporting_config.get("consent", False)


consent = HumbugConsent(consent_from_reporting_config_file)

session_id = str(uuid.uuid4())
bugout_reporting_config = get_reporting_config()
client_id = bugout_reporting_config.get("client_id")

hub_reporter = HumbugReporter(
    name="activeloopai/Hub",
    consent=consent,
    client_id=client_id,
    session_id=session_id,
    bugout_token=BUGOUT_TOKEN,
    tags=[],
)

hub_user = bugout_reporting_config.get("username")
if hub_user is not None:
    hub_reporter.tags.append(f"username:{hub_user}")

machine_id = bugout_reporting_config.get("machine_id")
if machine_id is not None:
    hub_reporter.tags.append(f"machine_id:{machine_id}")


def feature_report_path(
    path: str,
    feature_name: str,
    parameters: dict,
    starts_with: str = "hub://",
    token: str = None,
):
    """Helper function for generating humbug feature reports depending on the path"""
    if not isinstance(path, str):
        path = str(path)
    if path.startswith(starts_with):
        parameters["Path"] = path

    if token is not None:
        client = HubBackendClient(token=token)
        username = client.get_user_profile()["name"]

        index, current_username = find_current_username()

        if current_username is None:
            hub_reporter.tags.append(f"username:{username}")
        else:
            if f"username:{username}" != current_username:
                hub_reporter.tags[index] = f"username:{username}"

    hub_reporter.feature_report(
        feature_name=feature_name,
        parameters=parameters,
    )


def find_current_username():
    for index, tag in enumerate(hub_reporter.tags):
        if "username" in tag:
            return index, tag
    return None, None



import datetime
import posixpath
import pickle
import json
import os
import tempfile
import time
from typing import Dict, Optional, Tuple, Union

try:
    from google.cloud import storage  # type: ignore
    from google.api_core import retry  # type: ignore
    from google.oauth2 import service_account  # type: ignore
    import google.auth as gauth  # type: ignore
    import google.auth.compute_engine  # type: ignore
    import google.auth.credentials  # type: ignore
    import google.auth.exceptions  # type: ignore
    from google_auth_oauthlib.flow import InstalledAppFlow  # type: ignore
    from google.api_core.exceptions import NotFound  # type: ignore

    _GOOGLE_PACKAGES_INSTALLED = True
except ImportError:
    _GOOGLE_PACKAGES_INSTALLED = False


from hub.core.storage.provider import StorageProvider
from hub.util.exceptions import (
    GCSDefaultCredsNotFoundError,
    RenameError,
    PathNotEmptyException,
)
from hub.client.client import HubBackendClient


def _remove_protocol_from_path(path: str) -> str:
    return path.replace("gcp://", "").replace("gcs://", "").replace("gs://", "")


class GCloudCredentials:
    def __init__(self, token: Union[str, Dict] = None, project: str = None):
        self.scope = "https://www.googleapis.com/auth/cloud-platform"
        self.project = project
        self.credentials = None
        self.token = token
        self.tokens: Dict[str, Dict] = {}
        self.connect(method=token)

    def _load_tokens(self):
        """Get "browser" tokens from disc"""
        try:
            with open(".gcs_token", "rb") as f:
                return pickle.load(f)
        except Exception:
            return None

    def _save_tokens(self, token):
        with open(".gcs_token", "wb") as f:
            pickle.dump(token, f, 2)

    def _connect_google_default(self):
        """Attempts to get credentials from google default configurations:
            environment variable GOOGLE_APPLICATION_CREDENTIALS, Google Cloud SDK default credentials or default project.
            For more details see: https://google-auth.readthedocs.io/en/master/reference/google.auth.html#google.auth.default

        Raises:
            ValueError: If the name of the default project doesn't match the GCSProvider project name.
            DefaultCredentialsError: If no credentials are found.
        """
        credentials, project = gauth.default(scopes=[self.scope])
        if self.project and self.project != project:
            raise ValueError(
                "Project name does not match the google default project name."
            )
        self.project = project
        self.credentials = credentials

    def _connect_cache(self):
        """Load token stored after using _connect_browser() method."""
        credentials = self._load_tokens()
        if credentials:
            self.credentials = credentials

    def _dict_to_credentials(self, token: Dict):
        """
        Convert dict-style token.
        Does not preserve access token itself, assumes refresh required.

        Args:
            token (Dict): dictionary with token to be stored in .json file.

        Returns:
            Path to stored .json file.
        """
        token_file = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
        with open(token_file.name, "w") as f:
            json.dump(token, f)
        return token_file.name

    def _connect_token(self, token: Union[str, Dict] = None):
        """
        Connect using a concrete token.

        Args:
            token (str/dict/Credentials):
                If a str, try to load as a Service file, or next as a JSON; if
                dict, try to interpret as credentials; if Credentials, use directly.

        Raises:
            FileNotFoundError: If token file doesn't exist.
            ValueError: If token format isn't supported by gauth.
        """
        if isinstance(token, str):
            if not os.path.exists(token):
                raise FileNotFoundError(token)
            try:
                self._connect_service(token)
                return
            except:
                token = json.load(open(token))
        if isinstance(token, dict):
            token = self._dict_to_credentials(token)
            self._connect_service(token)
            return
        elif isinstance(token, google.auth.credentials.Credentials):
            credentials = token
        else:
            raise ValueError("Token format not understood")
        self.credentials = credentials

    def _connect_service(self, fn):
        credentials = service_account.Credentials.from_service_account_file(
            fn, scopes=[self.scope]
        )
        self.credentials = credentials

    def _connect_browser(self):
        """Create and store new credentials using OAuth authentication method.
            Requires having default client configuration file in ~/.config/gcloud/application_default_credentials.json
            (default location after initializing gcloud).

        Raises:
            GCSDefaultCredsNotFoundError: if application deafault credentials can't be found.
        """
        try:
            if os.name == "nt":
                path = os.path.join(
                    os.getenv("APPDATA"), "gcloud/application_default_credentials.json"
                )
            else:
                path = posixpath.expanduser(
                    "~/.config/gcloud/application_default_credentials.json"
                )
            with open(path, "r") as f:
                default_config = json.load(f)
        except:
            raise GCSDefaultCredsNotFoundError()
        client_config = {
            "installed": {
                "client_id": default_config["client_id"],
                "client_secret": default_config["client_secret"],
                "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                "token_uri": "https://accounts.google.com/o/oauth2/token",
            }
        }
        flow = InstalledAppFlow.from_client_config(client_config, [self.scope])
        credentials = flow.run_console()
        self._save_tokens(credentials)
        self.credentials = credentials

    def _connect_anon(self):
        """Use provider without specific credentials. Applicable for public projects/buckets."""
        self.credentials = None

    def connect(self, method: Union[str, Dict] = None):
        """
        Establish session token. A new token will be requested if the current
        one is within 100s of expiry.

        Args:
            method (str/Dict): Supported methods: ``google_default | cache | anon | browser | None``.
                Type of authorisation to implement - calls `_connect_*` methods.
                If None, will try sequence of methods.

        Raises:
            AttributeError: If method is invalid.
        """
        if method not in [
            "google_default",
            "cache",
            "anon",
            "browser",
            None,
        ]:
            self._connect_token(method)
        elif method is None:
            for meth in ["google_default", "cache", "anon"]:
                self.connect(method=meth)
                break
        elif isinstance(method, str):
            self.__getattribute__("_connect_" + method)()
        else:
            raise AttributeError(f"Invalid method: {method}")


class GCSProvider(StorageProvider):
    """Provider class for using GC storage."""

    def __init__(self, root: str, token: Union[str, Dict] = None, project: str = None):
        """Initializes the GCSProvider.

        Example:
            >>> gcs_provider = GCSProvider("gcs://my-bucket/gcs_ds")

        Args:
            root (str): The root of the provider. All read/write request keys will be appended to root.
            token (str/Dict): GCP token, used for fetching credentials for storage).
                Can be a path to the credentials file, actual credential dictionary or one of the folowing:
                - ``google_default``: Tries to load default credentials for the specified project.
                - ``cache``: Retrieves the previously used credentials from cache if exist.
                - ``anon``: Sets ``credentials=None``.
                - ``browser``: Generates and stores new token file using cli.
            project (str): Name of the project from GCloud.

        Raises:
            ModuleNotFoundError: If google cloud packages aren't installed.
        """
        if not _GOOGLE_PACKAGES_INSTALLED:
            raise ModuleNotFoundError(
                "Google cloud packages are not installed. Run `pip install hub[gcp]`."
            )
        self.root = root
        self.token: Union[str, Dict, None] = token
        self.tag: Optional[str] = None
        self.project = project
        self.missing_exceptions = (
            FileNotFoundError,
            IsADirectoryError,
            NotADirectoryError,
            AttributeError,
            NotFound,
        )
        self._initialize_provider()
        self._presigned_urls: Dict[str, Tuple[str, float]] = {}
        self.expiration: Optional[str] = None

    def subdir(self, path: str):
        return self.__class__(
            root=posixpath.join(self.root, path),
            token=self.token,
            project=self.project,
        )

    def _initialize_provider(self):
        self._set_bucket_and_path()
        if not self.token:
            self.token = None
        self.scoped_credentials = GCloudCredentials(self.token, project=self.project)
        self.retry = retry.Retry(deadline=60)
        self.client = storage.Client(credentials=self.scoped_credentials.credentials)
        self._client_bucket = None

    @property
    def client_bucket(self):
        if self._client_bucket is None:
            self._client_bucket = self.client.get_bucket(self.bucket)
        return self._client_bucket

    def _set_bucket_and_path(self):
        root = _remove_protocol_from_path(self.root)
        split_root = root.split("/", 1)
        self.bucket = split_root[0]
        if len(split_root) > 1:
            self.path = split_root[1]
        else:
            self.path = ""
        if not self.path.endswith("/"):
            self.path += "/"

    def _get_path_from_key(self, key):
        return posixpath.join(self.path, key)

    def _all_keys(self):
        self._blob_objects = self.client_bucket.list_blobs(prefix=self.path)
        return {posixpath.relpath(obj.name, self.path) for obj in self._blob_objects}

    def _set_hub_creds_info(self, hub_path: str, expiration: str):
        """Sets the tag and expiration of the credentials. These are only relevant to datasets using Hub storage.
        This info is used to fetch new credentials when the temporary 12 hour credentials expire.

        Args:
            hub_path (str): The hub cloud path to the dataset.
            expiration (str): The time at which the credentials expire.
        """
        self.hub_path = hub_path
        self.tag = hub_path[6:]  # removing the hub:// part from the path
        self.expiration = expiration

    def clear(self, prefix=""):
        """Remove all keys with given prefix below root - empties out mapping.

        Warning:
            Exercise caution!
        """
        self.check_readonly()
        path = posixpath.join(self.path, prefix) if prefix else self.path
        blob_objects = self.client_bucket.list_blobs(prefix=path)
        for blob in blob_objects:
            try:
                blob.delete()
            except Exception:
                pass

    def rename(self, root):
        """Rename root folder."""
        self.check_readonly()
        path = _remove_protocol_from_path(root)
        new_bucket, new_path = path.split("/", 1)
        if new_bucket != self.client_bucket.name:
            raise RenameError
        blob_objects = self.client_bucket.list_blobs(prefix=self.path)
        dest_objects = self.client_bucket.list_blobs(prefix=new_path)
        for blob in dest_objects:
            raise PathNotEmptyException(use_hub=False)
        for blob in blob_objects:
            new_key = "/".join([new_path, posixpath.relpath(blob.name, self.path)])
            self.client_bucket.rename_blob(blob, new_key)

        self.root = root
        self.path = new_path
        if not self.path.endswith("/"):
            self.path += "/"

    def __getitem__(self, key):
        """Retrieve data."""
        return self.get_bytes(key)

    def get_bytes(
        self,
        path: str,
        start_byte: Optional[int] = None,
        end_byte: Optional[int] = None,
    ):
        """Gets the object present at the path within the given byte range.

        Args:
            path (str): The path relative to the root of the provider.
            start_byte (int, optional): If only specific bytes starting from ``start_byte`` are required.
            end_byte (int, optional): If only specific bytes up to ``end_byte`` are required.

        Returns:
            bytes: The bytes of the object present at the path within the given byte range.

        Raises:
            InvalidBytesRequestedError: If ``start_byte`` > ``end_byte`` or ``start_byte`` < 0 or ``end_byte`` < 0.
            KeyError: If an object is not found at the path.
        """
        try:
            blob = self.client_bucket.get_blob(self._get_path_from_key(path))
            if end_byte is not None:
                end_byte -= 1
            return blob.download_as_bytes(
                retry=self.retry, start=start_byte, end=end_byte
            )
        except self.missing_exceptions:
            raise KeyError(path)

    def __setitem__(self, key, value):
        """Store value in key."""
        self.check_readonly()
        blob = self.client_bucket.blob(self._get_path_from_key(key))
        if isinstance(value, memoryview):
            value = value.tobytes()
        elif isinstance(value, bytearray):
            value = bytes(value)
        # if isinstance(value, memoryview) and (
        #     value.strides == (1,) and value.shape == (len(value.obj),)
        # ):
        #     value = value.obj
        # value = bytes(value)
        blob.upload_from_string(value, retry=self.retry)

    def __iter__(self):
        """Iterating over the structure."""
        yield from [f for f in self._all_keys() if not f.endswith("/")]

    def __len__(self):
        """Returns length of the structure."""
        return len(self._all_keys())

    def __delitem__(self, key):
        """Remove key."""
        self.check_readonly()
        blob = self.client_bucket.blob(self._get_path_from_key(key))
        try:
            blob.delete()
        except self.missing_exceptions:
            raise KeyError(key)

    def __contains__(self, key):
        """Checks if key exists in mapping."""
        stats = storage.Blob(
            bucket=self.client_bucket, name=self._get_path_from_key(key)
        ).exists(self.client_bucket.client)
        return stats

    def __getstate__(self):
        return (
            self.root,
            self.token,
            self.missing_exceptions,
            self.project,
            self.read_only,
        )

    def __setstate__(self, state):
        self.root = state[0]
        self.token = state[1]
        self.missing_exceptions = state[2]
        self.project = state[3]
        self.read_only = state[4]
        self._initialize_provider()

    def get_presigned_url(self, key, full=False):
        if full:
            root = _remove_protocol_from_path(key)
            split_root = root.split("/", 1)
            bucket = split_root[0]
            key = split_root[1] if len(split_root) > 1 else ""

            client_bucket = self.client.get_bucket(bucket)
        else:
            client_bucket = self.client_bucket

        url = None
        cached = self._presigned_urls.get(key)
        if cached:
            url, t_store = cached
            t_now = time.time()
            if t_now - t_store > 3200:
                del self._presigned_urls[key]
                url = None

        if url is None:
            if self._is_hub_path:
                client = HubBackendClient(self.token)  # type: ignore
                org_id, ds_name = self.tag.split("/")  # type: ignore
                url = client.get_presigned_url(org_id, ds_name, key)
            else:
                blob = client_bucket.get_blob(
                    self._get_path_from_key(key) if not full else key
                )
                url = blob.generate_signed_url(datetime.timedelta(seconds=3600))
            self._presigned_urls[key] = (url, time.time())
        return url

    def get_object_size(self, key: str) -> int:
        blob = self.client_bucket.get_blob(self._get_path_from_key(key))
        return blob.size

    def get_object_from_full_url(self, url: str):
        root = _remove_protocol_from_path(url)
        split_root = root.split("/", 1)
        bucket = split_root[0]
        path = split_root[1] if len(split_root) > 1 else ""

        client_bucket = self.client.get_bucket(bucket)

        try:
            blob = client_bucket.get_blob(path)
            return blob.download_as_bytes(retry=self.retry)
        except self.missing_exceptions:
            raise KeyError(path)



from collections import OrderedDict, defaultdict
from typing import List

from hub.util.hash import hash_str_to_int32
from hub.client.log import logger
import numpy as np
import hub


def convert_to_idx(samples, class_names: List[str]):
    class_idx = {class_names[i]: i for i in range(len(class_names))}

    def convert(samples):
        idxs = []
        additions = []
        for sample in samples:
            if isinstance(sample, np.ndarray):
                sample = sample.tolist()
            if isinstance(sample, str):
                idx = class_idx.get(sample)
                if idx is None:
                    idx = len(class_idx)
                    class_idx[sample] = idx
                    additions.append((sample, idx))
                idxs.append(idx)
            elif isinstance(sample, list):
                idxs_, additions_ = convert(sample)
                idxs.append(idxs_)
                additions.extend(additions_)
            else:
                idxs.append(sample)
        return idxs, additions

    return convert(samples)


def convert_to_hash(samples, hash_label_map):
    if isinstance(samples, np.ndarray):
        samples = samples.tolist()
    if isinstance(samples, list):
        return [convert_to_hash(sample, hash_label_map) for sample in samples]
    else:
        if isinstance(samples, str):
            hash_ = hash_str_to_int32(samples)
            hash_label_map[hash_] = samples
        else:
            hash_ = samples
        return hash_


def convert_hash_to_idx(hashes, hash_idx_map):
    if isinstance(hashes, list):
        return [convert_hash_to_idx(hash, hash_idx_map) for hash in hashes]
    else:
        try:
            return hash_idx_map[hashes]
        except KeyError:
            return hashes


def convert_to_text(inp, class_names: List[str]):
    if isinstance(inp, np.integer):
        idx = int(inp)
        if idx < len(class_names):
            return class_names[idx]
        return None
    return [convert_to_text(item, class_names) for item in inp]


def sync_labels(
    ds, label_temp_tensors, hash_label_maps, num_workers, scheduler, verbose=True
):
    hl_maps = defaultdict(OrderedDict)
    for map in hash_label_maps:
        for tensor in map:
            hl_maps[tensor].update(map[tensor])
    hash_label_maps = hl_maps

    @hub.compute
    def class_label_sync(
        hash_tensor_sample,
        samples_out,
        label_tensor: str,
        hash_idx_map,
    ):
        hashes = hash_tensor_sample.numpy().tolist()
        idxs = convert_hash_to_idx(hashes, hash_idx_map)
        samples_out[label_tensor].append(idxs)

    for tensor, temp_tensor in label_temp_tensors.items():
        if len(ds[temp_tensor]) == 0:
            ds.delete_tensor(temp_tensor, large_ok=True)
        else:
            try:
                target_tensor = ds[tensor]
                hash_label_map = hash_label_maps[temp_tensor]
                class_names = target_tensor.info.class_names
                new_labels = [
                    label
                    for label in hash_label_map.values()
                    if label not in class_names
                ]
                if verbose:
                    N = len(class_names)
                    for i in range(len(new_labels)):
                        logger.info(
                            f"'{new_labels[i]}' added to {tensor}.info.class_names at index {N + i}"
                        )
                class_names.extend(new_labels)
                label_idx_map = {class_names[i]: i for i in range(len(class_names))}
                hash_idx_map = {
                    hash: label_idx_map[hash_label_map[hash]] for hash in hash_label_map
                }
                target_tensor.info.is_dirty = True
                target_tensor.meta._disable_temp_transform = True
                target_tensor.meta.is_dirty = True

                logger.info("Synchronizing class labels...")
                class_label_sync(label_tensor=tensor, hash_idx_map=hash_idx_map).eval(
                    ds[temp_tensor],
                    ds,
                    num_workers=num_workers,
                    scheduler=scheduler,
                    progressbar=True,
                    check_lengths=False,
                    skip_ok=True,
                )
                target_tensor.meta._disable_temp_transform = False
            finally:
                ds.delete_tensor(temp_tensor, large_ok=True)



from typing import Optional


compression_ratios = {None: 1.0, "jpeg": 0.5, "png": 0.5, "webp": 0.5, "lz4": 0.5}


def get_compression_ratio(compression: Optional[str]) -> float:
    return compression_ratios.get(compression, 0.5)



r"""
The hub package provides a database which stores data as compressed chunked arrays that can be stored anywhere and 
later streamed to deep learning models.
"""

import threading
from queue import Queue
from botocore.config import Config
import numpy as np
import multiprocessing
import sys
from hub.util.check_latest_version import warn_if_update_required

if sys.platform == "darwin":
    multiprocessing.set_start_method("fork", force=True)


from .api.dataset import dataset as api_dataset
from .api.read import read
from .api.link import link
from .api.tiled import tiled
from .core.dataset import Dataset
from .core.transform import compute, compose
from .core.tensor import Tensor
from .util.bugout_reporter import hub_reporter
from .compression import SUPPORTED_COMPRESSIONS
from .htype import HTYPE_CONFIGURATIONS
from .htype import htype
from .integrations import huggingface
from .integrations import wandb

compressions = list(SUPPORTED_COMPRESSIONS)
htypes = sorted(list(HTYPE_CONFIGURATIONS))
list = api_dataset.list
exists = api_dataset.exists
load = api_dataset.load
empty = api_dataset.empty
like = api_dataset.like
delete = api_dataset.delete
rename = api_dataset.rename
copy = api_dataset.copy
deepcopy = api_dataset.deepcopy
ingest = api_dataset.ingest
ingest_kaggle = api_dataset.ingest_kaggle
ingest_dataframe = api_dataset.ingest_dataframe
ingest_huggingface = huggingface.ingest_huggingface
dataset = api_dataset.init  # type: ignore
tensor = Tensor

__all__ = [
    "tensor",
    "read",
    "link",
    "__version__",
    "load",
    "empty",
    "exists",
    "compute",
    "compose",
    "copy",
    "dataset",
    "Dataset",
    "deepcopy",
    "like",
    "list",
    "ingest",
    "ingest_kaggle",
    "ingest_huggingface",
    "compressions",
    "htypes",
    "config",
    "delete",
    "copy",
    "rename",
]

__version__ = "2.8.4"
warn_if_update_required(__version__)
__encoded_version__ = np.array(__version__)
config = {"s3": Config(max_pool_connections=50, connect_timeout=300, read_timeout=300)}


hub_reporter.tags.append(f"version:{__version__}")
hub_reporter.system_report(publish=True)
hub_reporter.setup_excepthook(publish=True)

event_queue: Queue = Queue()


def send_event():
    while True:
        try:
            event = event_queue.get()
            client, event_dict = event
            client.send_event(event_dict)
        except Exception:
            pass


threading.Thread(target=send_event, daemon=True).start()



import hub
import pytest
from hub.util.exceptions import DatasetHandlerError
from hub.util.access_method import parse_access_method


def test_access_method_parsing():
    assert parse_access_method("download") == ("download", 0, "threaded")
    assert parse_access_method("download:processed") == ("download", 0, "processed")
    assert parse_access_method("download:5") == ("download", 5, "threaded")
    assert parse_access_method("download:processed:5") == ("download", 5, "processed")
    assert parse_access_method("download:5:processed") == ("download", 5, "processed")
    with pytest.raises(ValueError):
        parse_access_method("download:5:processed:5")
    with pytest.raises(ValueError):
        parse_access_method("download:processed:processed")
    with pytest.raises(ValueError):
        parse_access_method("download:processed:5:processed")


def test_access_method(s3_ds_generator):
    with pytest.raises(DatasetHandlerError):
        hub.dataset("./some_non_existent_path", access_method="download")

    with pytest.raises(DatasetHandlerError):
        hub.dataset("./some_non_existent_path", access_method="local")

    ds = s3_ds_generator()
    with ds:
        ds.create_tensor("x")
        for i in range(10):
            ds.x.append(i)

    ds = s3_ds_generator(access_method="download")
    with pytest.raises(DatasetHandlerError):
        # second time download is not allowed
        s3_ds_generator(access_method="download")
    assert not ds.path.startswith("s3://")
    for i in range(10):
        assert ds.x[i].numpy() == i

    with pytest.raises(ValueError):
        s3_ds_generator(access_method="invalid")

    with pytest.raises(ValueError):
        s3_ds_generator(access_method="download", overwrite=True)

    with pytest.raises(ValueError):
        s3_ds_generator(access_method="local", overwrite=True)

    ds = s3_ds_generator(access_method="local")
    assert not ds.path.startswith("s3://")
    for i in range(10):
        assert ds.x[i].numpy() == i

    ds.delete()



from hub.core.storage.gcs import GCSProvider
from hub.util.storage import storage_provider_from_hub_path
from hub.core.storage.s3 import S3Provider
from hub.core.storage.google_drive import GDriveProvider
from hub.core.storage.local import LocalProvider
from hub.core.storage.memory import MemoryProvider
from hub.constants import (
    PYTEST_S3_PROVIDER_BASE_ROOT,
    PYTEST_GCS_PROVIDER_BASE_ROOT,
    S3_OPT,
    GCS_OPT,
)
from hub.tests.common import is_opt_true
import pytest


enabled_storages = pytest.mark.parametrize(
    "storage",
    ["memory_storage", "local_storage", "s3_storage", "gcs_storage", "gdrive_storage"],
    indirect=True,
)

enabled_persistent_storages = pytest.mark.parametrize(
    "storage",
    ["local_storage", "s3_storage", "gcs_storage", "gdrive_storage"],
    indirect=True,
)


enabled_remote_storages = pytest.mark.parametrize(
    "storage",
    [
        "s3_storage",
        "gcs_storage",
        "gdrive_storage",
        "gcs_root_storage",
        "s3_root_storage",
    ],
    indirect=True,
)


@pytest.fixture
def memory_storage(memory_path):
    return MemoryProvider(memory_path)


@pytest.fixture
def local_storage(local_path):
    return LocalProvider(local_path)


@pytest.fixture
def s3_storage(s3_path):
    return S3Provider(s3_path)


@pytest.fixture
def gdrive_storage(gdrive_path, gdrive_creds):
    return GDriveProvider(gdrive_path, token=gdrive_creds)


@pytest.fixture
def gcs_storage(gcs_path):
    return GCSProvider(gcs_path)


@pytest.fixture
def s3_root_storage(request):
    if not is_opt_true(request, GCS_OPT):
        pytest.skip()
        return

    return S3Provider(PYTEST_S3_PROVIDER_BASE_ROOT)


@pytest.fixture
def gcs_root_storage(request, gcs_creds):
    if not is_opt_true(request, GCS_OPT):
        pytest.skip()
        return

    return GCSProvider(PYTEST_GCS_PROVIDER_BASE_ROOT, token=gcs_creds)


@pytest.fixture
def hub_cloud_storage(hub_cloud_path, hub_cloud_dev_token):
    return storage_provider_from_hub_path(hub_cloud_path, token=hub_cloud_dev_token)


@pytest.fixture
def storage(request):
    """Used with parametrize to use all enabled storage fixtures."""
    return request.getfixturevalue(request.param)



from abc import ABC, abstractmethod
from hub.util.path import find_root
from pathlib import Path


class UnstructuredDataset(ABC):
    def __init__(self, source: str):
        self.source = Path(find_root(source))

    """Initializes an unstructured dataset.
    
    Args:
        source (str): The local path to folder containing an unstructured dataset and of the form ./path/to/dataset or ~/path/to/dataset or path/to/dataset.
    """

    @abstractmethod
    def structure(ds, use_progress_bar: bool = True, **kwargs):
        pass



from hub.constants import MB, PARTIAL_NUM_SAMPLES
from hub.core.chunk.sample_compressed_chunk import SampleCompressedChunk
import numpy as np
import pytest

import hub
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.sample import Sample  # type: ignore
from hub.core.tiling.deserialize import np_list_to_sample
from hub.core.tiling.sample_tiles import SampleTiles

compressions_paremetrized = pytest.mark.parametrize("compression", ["lz4"])


common_args = {
    "min_chunk_size": 1 * MB,
    "max_chunk_size": 2 * MB,
    "tiling_threshold": 1 * MB,
}


def create_tensor_meta():
    tensor_meta = TensorMeta()
    tensor_meta.dtype = "float32"
    tensor_meta.max_shape = None
    tensor_meta.min_shape = None
    tensor_meta.htype = None
    tensor_meta.length = 0
    return tensor_meta


@compressions_paremetrized
def test_read_write_sequence(compression):
    tensor_meta = create_tensor_meta()
    common_args["tensor_meta"] = tensor_meta
    common_args["compression"] = compression
    dtype = tensor_meta.dtype
    data_in = [np.random.rand(250, 125, 3).astype(dtype) for _ in range(10)]
    data_in2 = data_in.copy()
    while data_in:
        chunk = SampleCompressedChunk(**common_args)
        num_samples = int(chunk.extend_if_has_space(data_in))
        data_out = [chunk.read_sample(i) for i in range(num_samples)]
        np.testing.assert_array_equal(data_out, data_in2[:num_samples])
        data_in = data_in[num_samples:]
        data_in2 = data_in2[num_samples:]


@compressions_paremetrized
def test_read_write_sequence_big(cat_path, compression):
    tensor_meta = create_tensor_meta()
    common_args = {
        "min_chunk_size": 16 * MB,
        "max_chunk_size": 32 * MB,
        "tiling_threshold": 16 * MB,
        "tensor_meta": tensor_meta,
        "compression": compression,
    }

    dtype = tensor_meta.dtype
    data_in = []
    for i in range(50):
        if i % 10 == 0:
            data_in.append(np.random.rand(6001, 3000, 3).astype(dtype))
        elif i % 3 == 0:
            data_in.append(hub.read(cat_path))
        else:
            data_in.append(np.random.rand(1000, 500, 3).astype(dtype))
    data_in2 = data_in.copy()
    tiles = []
    original_length = len(data_in)

    while data_in:
        chunk = SampleCompressedChunk(**common_args)
        num_samples = chunk.extend_if_has_space(data_in)
        if num_samples == PARTIAL_NUM_SAMPLES:
            tiles.append(chunk.read_sample(0, is_tile=True))
            sample = data_in[0]
            assert isinstance(sample, SampleTiles)
            if sample.is_last_write:
                current_length = len(data_in)
                index = original_length - current_length
                full_data_out = np_list_to_sample(
                    tiles,
                    sample.sample_shape,
                    sample.tile_shape,
                    sample.layout_shape,
                    dtype,
                )
                np.testing.assert_array_equal(full_data_out, data_in2[index])
                data_in = data_in[1:]
                tiles = []

        elif num_samples > 0:
            data_out = [chunk.read_sample(i) for i in range(num_samples)]
            for i, item in enumerate(data_out):
                if isinstance(item, Sample):
                    item = item.array
                np.testing.assert_array_equal(item, data_in[i])
            data_in = data_in[num_samples:]


# @compressions_paremetrized
# def test_update(compression):
#     tensor_meta = create_tensor_meta()
#     common_args["tensor_meta"] = tensor_meta
#     common_args["compression"] = compression
#     dtype = tensor_meta.dtype
#     arr = np.random.rand(7, 25, 125, 3).astype(dtype)
#     data_in = list(arr)
#     chunk = SampleCompressedChunk(**common_args)
#     chunk.extend_if_has_space(data_in)
#     data_out = np.array([chunk.read_sample(i) for i in range(7)])
#     np.testing.assert_array_equal(data_out, data_in)

#     data_3 = np.random.rand(175, 350, 3).astype(dtype)
#     data_5 = np.random.rand(1500, 750, 3).astype(dtype)

#     chunk.update_sample(3, data_3)
#     chunk.update_sample(5, data_5)
#     for i in range(7):
#         if i == 3:
#             np.testing.assert_array_equal(chunk.read_sample(i), data_3)
#         elif i == 5:
#             np.testing.assert_array_equal(chunk.read_sample(i), data_5)
#         else:
#             np.testing.assert_array_equal(chunk.read_sample(i), arr[i])



from typing import Any, Dict
from hub.core.meta.meta import Meta
from hub.core.index import Index
import posixpath


class DatasetMeta(Meta):
    def __init__(self):
        super().__init__()
        self.tensors = []
        self.groups = []
        self.tensor_names = {}
        self.hidden_tensors = []
        self.default_index = Index().to_json()

    @property
    def visible_tensors(self):
        return list(
            filter(
                lambda t: self.tensor_names[t] not in self.hidden_tensors,
                self.tensor_names.keys(),
            )
        )

    @property
    def nbytes(self):
        # TODO: can optimize this
        return len(self.tobytes())

    def __getstate__(self) -> Dict[str, Any]:
        d = super().__getstate__()
        d["tensors"] = self.tensors
        d["groups"] = self.groups
        d["tensor_names"] = self.tensor_names
        d["hidden_tensors"] = self.hidden_tensors
        d["default_index"] = self.default_index
        return d

    def add_tensor(self, name, key, hidden=False):
        if key not in self.tensors:
            self.tensor_names[name] = key
            self.tensors.append(key)
            if hidden:
                self.hidden_tensors.append(key)
            self.is_dirty = True

    def _hide_tensor(self, name):
        assert name in self.tensor_names
        if name not in self.hidden_tensors:
            self.hidden_tensors.append(self.tensor_names[name])
            self.is_dirty = True

    def add_group(self, name):
        if name not in self.groups:
            self.groups.append(name)
            self.is_dirty = True

    def delete_tensor(self, name):
        key = self.tensor_names.pop(name)
        self.tensors.remove(key)
        self.is_dirty = True

    def delete_group(self, name):
        self.groups = list(filter(lambda g: not g.startswith(name), self.groups))
        self.tensors = list(filter(lambda t: not t.startswith(name), self.tensors))
        self.hidden_tensors = list(
            filter(lambda t: not t.startswith(name), self.hidden_tensors)
        )
        tensor_names_keys = list(self.tensor_names.keys())
        for key in tensor_names_keys:
            if key.startswith(name):
                self.tensor_names.pop(key)
        self.is_dirty = True

    def rename_tensor(self, name, new_name):
        key = self.tensor_names.pop(name)
        self.tensor_names[new_name] = key
        self.is_dirty = True

    def rename_group(self, name, new_name):
        self.groups.remove(name)
        self.groups = list(
            map(
                lambda g: posixpath.join(new_name, posixpath.relpath(g, name))
                if (g == name or g.startswith(name + "/"))
                else g,
                self.groups,
            )
        )
        self.groups.append(new_name)
        self.is_dirty = True



from hub.core.compute.provider import ComputeProvider


class SerialProvider(ComputeProvider):
    def __init__(self):
        pass

    def map(self, func, iterable):
        return list(map(func, iterable))

    def map_with_progressbar(self, func, iterable, total_length: int, desc=None):
        from tqdm.std import tqdm  # type: ignore

        pbar = tqdm(total=total_length, desc=desc)

        def sub_func(*args, **kwargs):
            def pg_callback(value: int):
                pbar.update(value)

            return func(pg_callback, *args, **kwargs)

        result = self.map(sub_func, iterable)

        return result

    def create_queue(self):
        raise NotImplementedError("no queues in serial provider")

    def close(self):
        return



from typing import Optional
import hub
from hub.core.storage.provider import StorageProvider
from hub.core.storage.lru_cache import LRUCache
from hub.core.storage import MemoryProvider
from hub.constants import MB


def remove_memory_cache(storage: StorageProvider):
    """Removes the memory cache."""
    if isinstance(storage, LRUCache) and isinstance(
        storage.cache_storage, MemoryProvider
    ):
        return storage.next_storage
    return storage


def get_base_storage(storage: StorageProvider):
    """Removes all layers of caching and returns the underlying storage."""
    while isinstance(storage, LRUCache):
        if storage.next_storage is not None:
            storage = storage.next_storage
        else:
            storage = storage.cache_storage
    return storage


def get_dataset_with_zero_size_cache(ds):
    """Returns a dataset with same storage but cache size set to zero."""
    ds_base_storage = get_base_storage(ds.storage)
    zero_cache_storage = LRUCache(MemoryProvider(), ds_base_storage, 0)
    commit_id = ds.pending_commit_id
    ds = hub.core.dataset.dataset_factory(
        path=ds.path,
        storage=zero_cache_storage,
        index=ds.index,
        group_index=ds.group_index,
        read_only=ds.read_only,
        token=ds.token,
        verbose=False,
        link_creds=ds.link_creds,
        pad_tensors=ds._pad_tensors,
    )
    if ds.pending_commit_id != commit_id:
        ds.checkout(commit_id)
    return ds


def create_read_copy_dataset(dataset, commit_id: Optional[str] = None):
    """Creates a read-only copy of the given dataset object, without copying underlying data.

    Args:
        dataset: The Dataset object to copy.
        commit_id: The commit id to checkout the new read-only copy to.

    Returns:
        A new Dataset object in read-only mode.
    """
    base_storage = get_base_storage(dataset.storage)
    if isinstance(base_storage, MemoryProvider):
        new_storage = base_storage.copy()
        new_storage.dict = base_storage.dict
    else:
        new_storage = base_storage.copy()
    storage = LRUCache(MemoryProvider(), new_storage, 256 * MB)
    ds = dataset.__class__(
        storage,
        index=dataset.index,
        group_index=dataset.group_index,
        read_only=True,
        public=dataset.public,
        token=dataset._token,
        verbose=False,
        path=dataset.path,
    )
    if commit_id is not None:
        ds.checkout(commit_id)
    return ds
